<?php

namespace App\Core\Elastic\Category;

use App\Core\Elastic\StagedIndex;
use Elasticsearch\Client;

class CategoryIndex extends StagedIndex
{
    protected $name = 'categories';
    
    /** @var Client */
    private $client;
    
    /**
     * ProductIndex constructor.
     * @param Client $client
     */
    public function __construct(Client $client)
    {
        $this->client = $client;
    }
    
    public function down(): void
    {
        $this->client->indices()->delete(['index' => $this->fullIndexName()]);
    }
    
    public function up(): void
    {
        $this->client->indices()->create([
            'index' => $this->fullIndexName(),
            'body' => [
                'settings' => [
                    'number_of_shards' => 1,
                    'index.number_of_replicas' => '0',
                    'analysis' => [
                        "char_filter"=> [
                            "rus_en_key" => [
                                "type" => "mapping",
                                "mappings" => [
                                    "a => ф",
                                    "b => и",
                                    "c => с",
                                    "d => в",
                                    "e => у",
                                    "f => а",
                                    "g => п",
                                    "h => р",
                                    "i => ш",
                                    "j => о",
                                    "k => л",
                                    "l => д",
                                    "m => ь",
                                    "n => т",
                                    "o => щ",
                                    "p => з",
                                    "r => к",
                                    "s => ы",
                                    "t => е",
                                    "u => г",
                                    "v => м",
                                    "w => ц",
                                    "x => ч",
                                    "y => н",
                                    "z => я",
                                    "A => Ф",
                                    "B => И",
                                    "C => С",
                                    "D => В",
                                    "E => У",
                                    "F => А",
                                    "G => П",
                                    "H => Р",
                                    "I => Ш",
                                    "J => О",
                                    "K => Л",
                                    "L => Д",
                                    "M => Ь",
                                    "N => Т",
                                    "O => Щ",
                                    "P => З",
                                    "R => К",
                                    "S => Ы",
                                    "T => Е",
                                    "U => Г",
                                    "V => М",
                                    "W => Ц",
                                    "X => Ч",
                                    "Y => Н",
                                    "Z => Я",
                                    "[ => х",
                                    "] => ъ",
                                    "; => ж",
                                    "< => б",
                                    "> => ю",
                                ],
                            ],
                        ],
                        'filter' => [
                            'russian_stop' => [
                                'type' => 'stop',
                                'stopwords' => '_russian_',
                            ],
                            'russian_stemmer' => [
                                'type' => 'stemmer',
                                'language' => 'russian',
                            ],
                        ],
                        'analyzer' => [
                            'default' => [
                                'type' => 'custom',
                                "char_filter" => ['html_strip', 'rus_en_key'],
                                'tokenizer' => 'standard',
                                'filter' => ['lowercase', 'russian_stop', 'russian_stemmer'],
                            ],
                        ],
                    ],
                ],
                'mappings' => [
                    'properties' => [
                        'name' => [
                            'type' => 'text',
                        ],
                    ],
                ],
            ],
        ]);
    }
    
    public function isUp(): bool
    {
        return $this->client->indices()->exists(['index' => $this->fullIndexName()]);
    }
    
    /**
     * Добавить товар в индекс
     * @param int $id
     * @param string $name
     */
    public function upsert(int $id, string $name): void
    {
        $this->client->index([
            'index' => $this->fullIndexName(),
            'id' => $id,
            'body' => [
                'name' => $name,
            ],
        ]);
    }
    
    /**
     * Убрать товар из индекса
     * @param int $id
     */
    public function delete(int $id): void
    {
        $this->client->delete([
            'index' => $this->fullIndexName(),
            'id' => $id,
        ]);
    }
    
    public function search(string $query, int $limit)
    {
        $response = $this->client->search([
            'index' => $this->fullIndexName(),
            'body' => [
                'size' => $limit,
                'query' => [
                    'match' => [
                        'name' => [
                            'query' => $query,
                            'fuzziness' => 0,
                        ],
                    ],
                ],
            ],
        ]);
        
        $result = [];
        if ($response['hits']['hits']) {
            foreach ($response['hits']['hits'] as $hit) {
                $result[] = (int)$hit['_id'];
            }
        }

        return $result;
    }
}
