<?php

namespace App\Core\Elastic;

abstract class ElasticQuery
{
    const MAX_LIMIT = 100000;

    protected $mustContext = [];
    protected $shouldContext = [];
    protected $filterContext = [];

    /** @var int */
    private $limit = self::MAX_LIMIT;
    /** @var int */
    private $offset = 0;

    private $order = [];

    public function query(): array
    {
        $query = [];
        if ($this->mustContext || $this->shouldContext || $this->filterContext) {
            if ($this->mustContext) {
                $query['bool']['must'] = $this->mustContext;
            }
            if ($this->shouldContext) {
                $query['bool']['should'] = $this->shouldContext;
            }
            if ($this->filterContext) {
                $query['bool']['filter'] = $this->filterContext;
            }
        } else {
            $query = [
                'match_all' => ['boost' => 1],
            ];
        }

        $body = [
            'query' => $query,
        ];

        if (!is_null($this->limit)) {
            $body['size'] = $this->limit;
        }
        if (!is_null($this->offset)) {
            $body['from'] = $this->offset;
        }

        if ($this->order) {
            $body['sort'] = [];
            foreach ($this->order as $field => $orderParams) {
                $body['sort'][] = [
                    $field => $orderParams,
                ];
            }
        }

        return $body;
    }

    public function page(int $number, int $size)
    {
        $this->offset = ($number - 1) * $size;
        $this->limit = $size;
    }

    public function pageOffset($offset, $limit)
    {
        $this->offset = $offset;
        $this->limit = $limit;
    }

    public function id($id)
    {
        $this->keyword('_id', $id);
    }

    public function orderBy($field, $direction = 'asc', $nested = [])
    {
        $this->order[$field] = [
            'order' => $direction,
        ];
        if ($nested) {
            $this->order[$field]['nested'] = $nested;
        }
    }

    protected function keyword($field, $value)
    {
        if (is_array($value)) {
            $condition = ['bool' => ['should' => []]];
            foreach ($value as $v) {
                $condition['bool']['should'][] = [
                    'term' => [$field => $v],
                ];
            }
            $this->filterContext[] = $condition;
        } else {
            $this->filterContext[] = [
                'term' => [$field => $value],
            ];
        }
    }

    protected function nestedKeyword($path, $field, $value)
    {
        if (is_array($value)) {
            $condition = ['bool' => ['should' => []]];
            foreach ($value as $v) {
                $condition['bool']['should'][] = [
                    'term' => [$field => $v],
                ];
            }
        } else {
            $condition = [
                'term' => [$field => $value],
            ];
        }

        $this->addNestedToFilterContext($path, $condition);
    }

    protected function fuzzyText($field, $value)
    {
        $this->mustContext[] = [
            'match' => [
                $field => [
                    'query' => $value,
                    'fuzziness' => 0,
                ],
            ],
        ];
    }

    protected function multiMatch($fields, $value)
    {
        $boostStep = 20;
        $boost = $boostStep + count($fields) * $boostStep;
        $this->mustContext[] = [
            'multi_match' => [
                'query' => $value,
                'fields' => $fields,
                'fuzziness' => 0,
                "operator" => "and",
                "boost"    => $boost + 50,
            ],
        ];

        foreach ($fields as $field) {
            $this->shouldContext[] = [
                'match' => [
                    $field => [
                        "query"    => $value,
                        'fuzziness' => 0,
                        "operator" => "or",
                        "boost"    => $boost,
                    ],
                ],
            ];
            $boost -= $boostStep;
        }
    }

    protected function range($field, ?string $after, ?string $before)
    {
        $condition = [
            'range' => [
                $field => [],
            ],
        ];
        if ($after) {
            $condition['range'][$field]['gte'] = $after;
        }

        if ($before) {
            $condition['range'][$field]['lte'] = $before;
        }

        $this->filterContext[] = $condition;
    }

    protected function nestedRange($path, $field, ?string $after, ?string $before)
    {
        $condition = [
            'range' => [
                $field => [],
            ],
        ];
        if ($after) {
            $condition['range'][$field]['gte'] = $after;
        }

        if ($before) {
            $condition['range'][$field]['lte'] = $before;
        }

        $this->addNestedToFilterContext($path, $condition);
    }

    protected function addNestedToFilterContext($path, $condition)
    {
        $foundNested = false;
        foreach ($this->filterContext as $k => $filter) {
            if (isset($filter['nested']['path']) && $filter['nested']['path'] == $path) {
                if (!isset($filter['nested']['query']['bool']['filter'])) {
                    $oldFilter = $filter['nested']['query'];
                    unset($this->filterContext[$k]['nested']['query']);
                    $this->filterContext[$k]['nested']['query']['bool']['filter'][] = $oldFilter;
                }
                $this->filterContext[$k]['nested']['query']['bool']['filter'][] = $condition;
                $foundNested = true;

                break;
            }
        }
        if (!$foundNested) {
            $this->filterContext[] = [
                'nested' => [
                    'path' => $path,
                    'query' => $condition,
                ],
            ];
        }
    }
}
