<?php

namespace App\Core\Elastic\Product\Indexing\Actions;

use App\Core\Elastic\Product\Indexing\OfferTicket;
use Ensi\MarketingClient\Api\DiscountsApi;
use Ensi\MarketingClient\Dto\CalculateForCatalogRequest;
use Ensi\MarketingClient\Dto\DiscountOfferCatalogRequest;
use Ensi\MarketingClient\Dto\OffersDiscountCatalogResponse;
use Illuminate\Support\Collection;

class FillPricesAction
{
    public function __construct(private DiscountsApi $api)
    {
    }

    /**
     * @param Collection|OfferTicket[] $tickets
     */
    public function execute(Collection $tickets): void
    {
        $prices = $this->loadPrices($tickets);

        foreach ($tickets as $ticket) {
            $price = $prices->get($ticket->id);

            if (!$price) {
                $ticket->target->addStrike('Нет цены');
                continue;
            }

            $ticket->target->setPrice([
                'cost' => $price->getCost(),
                'price' => $price->getPrice(),
            ]);
        }
    }

    /**
     * @param Collection|OfferTicket[] $tickets
     * @return Collection|OffersDiscountCatalogResponse[]
     */
    private function loadPrices(Collection $tickets): Collection
    {
        $offerRequests = $tickets->filter(fn(OfferTicket $ticket) => $ticket->hasPrice())
            ->map(fn(OfferTicket $ticket) => $this->createOfferRequest($ticket))
            ->values()
            ->all();

        if (!$offerRequests) {
            return new Collection();
        }

        $request = new CalculateForCatalogRequest();
        $request->setOffers($offerRequests);

        return collect($this->api->calculateForCatalog($request)?->getData() ?? [])
            ->keyBy(fn(OffersDiscountCatalogResponse $response) => $response->getOfferId());
    }

    private function createOfferRequest(OfferTicket $ticket): DiscountOfferCatalogRequest
    {
        return (new DiscountOfferCatalogRequest())
            ->setOfferId($ticket->id)
            ->setProductId($ticket->product->id)
            ->setBrandId($ticket->product->brand_id)
            ->setCategoryId($ticket->product->category_id)
            ->setPrice($ticket->price());
    }
}
