<?php

namespace App\Core\Elastic\Product;

/**
 * Класс записи с документом товара в соответствующем индексе Эластика.
 *
 * Class ProductIndexRecord
 * @package App\Core\Elastic\Product
 */
class ProductIndexRecord
{
    public $tags = [];

    /** @var bool Активность товара */
    private $active = true;
    /** @var int */
    public $archive;
    /** @var string Название товара */
    public $name;
    /** @var string Код товара */
    public $code;
    /** @var string Артикул */
    public $barcode;
    /** @var string Внутренний уникальный идентификатор (на витрине выводится как артикул) */
    public $externalId;
    /** @var mixed Дата добавления */
    public $dateAdd;

    /** @var bool */
    public $isNew;

    /** @var array Свойства */
    public $props = [];
    /** @var array Группы */
    public $groups = [];

    /** @var string Код бренда */
    public $brandCode;
    /** @var string Название бренда */
    public $brandName;

    /** @var string Название производителя */
    public $manufacturerName;
    /** @var string Название страны производства */
    public $countryName;

    /** @var string */
    public $productType;

    // коды категорий, перечисление от рутовой к нижней
    public $categoryCodes = [];
    // название актуальной категории
    public $categoryName;

    public $catalogImage;

    public $media = [];
    public $variantGroupId;

    public $weight;

    private $description;

    /** @var array|OfferIndexProperty[] Офферы */
    private $offers = [];

    /** @var array Ассоциативный массив характеристик товара */
    private $characteristics = [];

    private $strikes = [];

    /**
     * Добавляет вспомогательное сообщение о том, что какие-то данные не заполнены, некорректны и т.п.
     * Может в дальнейшем пригодиться в админке.
     * Страйк автоматически делает товар неактивным (скрывает его с витрины).
     */
    public function addStrike(string $msg): void
    {
        $this->active = false;
        $this->strikes[] = $msg;
    }

    public function addOfferStrike(int $sellerId, string $msg): void
    {
        $this->offer($sellerId)->addStrike($msg);
    }

    public function resetSaleActive(int $sellerId): void
    {
        $this->offer($sellerId)->resetSaleActive();
    }

    public function description(?string $text, $image, $video): void
    {
        $this->description = [
            'text' => $text,
            'image' => $image,
            'video' => $video,
        ];
    }

    public function addCharacteristic(string $name, mixed $value): void
    {
        $this->characteristics[] = [
            'name' => $name,
            'value' => $value,
        ];
    }

    public function addPrice(int $sellerId, $price): void
    {
        $this->offer($sellerId)->setPrice($price);
    }

    public function offer(int $offerId): OfferIndexProperty
    {
        if (isset($this->offers[$offerId])) {
            return $this->offers[$offerId];
        }

        $offer = new OfferIndexProperty();
        $offer->offerId = $offerId;
        $this->offers[$offerId] = $offer;

        return $offer;
    }

    /**
     * @return array
     */
    public function toArray(): array
    {
        return [
            'active' => $this->active,
            'archive' => $this->archive,

            'has_active_offers' => $this->hasActiveOffers(),

            'name' => $this->name,
            'code' => $this->code,
            'barcode' => $this->barcode,
            'external_id' => $this->externalId,
            'date_add' => $this->dateAdd,

            'offers' => $this->offersToArray(),

            'brand' => [
                'code' => $this->brandCode,
                'name' => $this->brandName,
            ],
            'category' => [
                'code' => $this->categoryCodes,
                'name' => $this->categoryName,
            ],
            'tags' => $this->tags,
            'props' => $this->props,
            'product_groups' => $this->groups,

            'catalog_image' => $this->catalogImage,

            'description' => $this->description,
            'media' => $this->media,
            'variant_group' => $this->variantGroupId,
            'characteristics' => $this->characteristics,
            'strikes' => $this->strikes,
        ];
    }

    private function offersToArray(): array
    {
        $result = [];

        foreach ($this->offers as $offer) {
            $result[] = $offer->toArray();
        }

        return $result;
    }

    /**
     * У товара есть хотя бы один активный оффер?
     *
     * @return bool
     */
    private function hasActiveOffers(): bool
    {
        foreach ($this->offers as $offer) {
            if ($offer->isActive()) {
                return true;
            }
        }

        return false;
    }
}
