<?php

namespace App\Core\Elastic\Product;

use App\Core\Elastic\ElasticQuery;

/**
 * Класс запроса в Эластик по товарам.
 *
 * Class ProductQuery
 * @package App\Core\Elastic\Product
 */
class ProductQuery extends ElasticQuery
{
    public const CATEGORY_COUNT_AGGREGATION = "categoryCounts";

    protected $aggregations = [];

    /**
     * @return array
     */
    public function query(): array
    {
        $rawQuery = parent::query();
        if ($this->aggregations) {
            $rawQuery["aggs"] = $this->aggregations;
        }

        return $rawQuery;
    }

    public function aggregateCategoryCounts()
    {
        $this->aggregations[self::CATEGORY_COUNT_AGGREGATION] = [
            "terms" => ["field" => "category.code", "size" => 100000],
        ];
    }

    public function code($code)
    {
        $this->keyword("code", $code);
    }

    public function category($code)
    {
        $this->keyword("category.code", $code);
    }

    public function active(bool $value)
    {
        $this->productActive($value);
        $this->offerActive($value);
    }

    public function productActive(bool $value)
    {
        $this->keyword("active", $value);
    }

    public function offerActive(bool $value)
    {
        $this->nestedKeyword("offers", "offers.active", $value);
    }

    public function seller_id($seller_id)
    {
        $this->nestedKeyword("offers", "offers.seller_id", $seller_id);
    }

    public function offer_id($offer_id)
    {
        $this->nestedKeyword("offers", "offers.offer_id", $offer_id);
    }

    public function saleActive(bool $value)
    {
        $this->nestedKeyword("offers", "offers.saleActive", $value);
    }

    /**
     * У товара есть хотя бы один активный оффер?
     *
     * @param bool $value
     */
    public function hasActiveOffers(bool $value)
    {
        $this->keyword("has_active_offers", $value);
    }

    public function price($from, $to)
    {
        $this->nestedRange("offers", "offers.price.price", $from, $to);
    }

    public function qty($from, $to)
    {
        $this->nestedRange("offers", "offers.qty", $from, $to);
    }

    /**
     * Ограничивает выборку товарами бренда code.
     *
     * @param $code
     */
    public function brand($code)
    {
        $this->keyword("brand.code", $code);
    }

    /**
     * Создан от after до before.
     *
     * @param $after
     * @param $before
     */
    public function created($after, $before)
    {
        $this->range("date_add", $after, $before);
    }

    /**
     * Значение архивности value.
     *
     * @param int $value
     */
    public function archive(int $value)
    {
        $this->keyword("archive", $value);
    }

    /**
     * Fuzzy-поиск по имени.
     *
     * @param string $value
     */
    public function name(string $value)
    {
        $this->fuzzyText("name", $value);
    }

    public function barcode($code)
    {
        $this->keyword("barcode", $code);
    }

    /**
     * Соответствует запросу полнотекстового поиска.
     *
     * @param $fullText
     */
    public function fullText($fullText)
    {
        $this->multiMatch(["name", "brand.name", "category.name", "external_id"], $fullText);
    }

    public function new()
    {
        $this->keyword("tags", ProductIndex::TAG_NEW);
    }

    public function productGroup($groupCode)
    {
        $this->keyword("product_groups", $groupCode);
    }

    public function property($name, $values)
    {
        $this->keyword("props", array_map(function ($value) use ($name) {
            return "{$name}#{$value}";
        }, $values));
    }

    public function orderByPrice($seller_id, $offer_id, $direction)
    {
        $nested = [
            'path' => 'offers',
        ];
        if ($seller_id) {
            if (is_array($seller_id)) {
                foreach ($seller_id as $v) {
                    $nested['filter']['bool']['should'][] = [
                        'term' => ['offers.seller_id' => $v],
                    ];
                }
            } else {
                $nested['filter'] = [
                    'term' => ['offers.seller_id' => $seller_id],
                ];
            }
        } elseif ($offer_id) {
            if (is_array($offer_id)) {
                foreach ($offer_id as $v) {
                    $nested['filter']['bool']['should'][] = [
                        'term' => ['offers.offer_id' => $v],
                    ];
                }
            } else {
                $nested['filter'] = [
                    'term' => ['offers.offer_id' => $offer_id],
                ];
            }
        }
        $this->orderBy("offers.price.price", $direction, $nested);
    }
}
