<?php

namespace App\Core\Rest\Eloquent;

use App\Core\Rest\RestQuery;
use App\Core\Rest\RestSerializable;
use Exception;
use Illuminate\Database\Eloquent\Builder;
use Illuminate\Database\Eloquent\Concerns\HasTimestamps;
use Illuminate\Database\Eloquent\Model;
use Illuminate\Database\Eloquent\Relations\Relation;
use Illuminate\Support\Facades\Schema;
use ReflectionClass;

abstract class AbstractRestModel extends Model implements RestSerializable
{
    use HasTimestamps;
    
    /**
     * @var bool
     */
    public static $snakeAttributes = false;
   
    /**
     * @var array - ['relation1', 'relation2', ..., 'relationN' => 'foreign_key']
     */
    protected static $restIncludes = [];
   
    /**
     * @return string
     */
    public static function restEntity(): string
    {
        try {
            $entity = (new ReflectionClass(static::class))
               ->getShortName();
        } catch (Exception $e) {
            $entity = end(explode('\\', static::class));
        }
        $entity = mb_strtolower($entity);
       
        return $entity;
    }
   
    /**
     * @param  Builder  $query
     * @param  RestQuery  $restQuery
     * @return Builder
     */
    public static function modifyQuery(Builder $query, RestQuery $restQuery): Builder
    {
        //Выбираемые поля сущности
        $fields = $restQuery->getFields(static::restEntity());
        if ($fields) {
            $query->select($fields);
        }
       
        //Получение данных по связанным сущностям (c учётом необходимых полей)
        if (static::$restIncludes) {
            foreach (static::$restIncludes as $key => $include) {
                $foreignFields = null;
                if (!is_numeric($key)) {
                    $foreignFields = $include;
                    $include = $key;
                }
               
                $includeFields = $restQuery->getFields($include);
                $includeWithFields = function (Builder $query, string $include, array $includeFields) {
                    $query->with([
                       $include => function (Relation $query) use ($includeFields) {
                           $query->select($includeFields);
                       },
                   ]);
                };
               
                if ($restQuery->isIncluded($include)) {
                    $foreignFields = is_null($foreignFields) ? $include . '_id' : $foreignFields;
                    $foreignFields = is_array($foreignFields) ? $foreignFields : [$foreignFields];
                   
                    foreach ($foreignFields as $foreignField) {
                        if (!in_array($foreignField, $fields) &&
                           Schema::hasColumn((new static())->getTable(), $foreignField)
                       ) {
                            if ($fields) {
                                $query->addSelect($foreignField);
                            } else {
                                $query->select(['*']);
                            }
                        }
                    }
                   
                    if ($includeFields) {
                        $includeWithFields($query, $include, $includeFields);
                    } else {
                        $query->with($include);
                    }
                } else {
                    //Выбираемые поля связанных с сущностей
                    if ($includeFields) {
                        $includeWithFields($query, $include, $includeFields);
                    }
                }
            }
        }
       
        //Фильтрация по полям сущности
        foreach ($restQuery->filterIterator() as [$field, $op, $value]) {
            if (($op == '=' && is_array($value)) || $op == 'in') {
                $query->whereIn($field, $value);
            } elseif ($op == 'null') {
                $query->whereNull($field);
            } elseif ($op == 'notNull') {
                $query->whereNotNull($field);
            } else {
                $query->where($field, $op, $value);
            }
        }
   
        //Сортировка по полям сущности
        foreach ($restQuery->sortIterator() as [$field, $dir]) {
            $query->orderBy($field, $dir);
        }
       
        return $query;
    }
   
    /**
     * @param  RestQuery  $restQuery
     * @return array
     */
    public function toRest(RestQuery $restQuery): array
    {
        return $this->toArray();
    }
}
