<?php

namespace App\Core\Rest;

use Generator;
use Illuminate\Http\Request;

/**
 * Class RestQuery - обёртка для формирования и получения параметров запроса в соотсветсвии с рекомендациями JSON:API
 * @package Common\Rest
 */
class RestQuery
{
    /**
     * @var array параметр json-api запроса include
     */
    private $includes = [];
    /**
     * @var array параметр json-api запроса fields
     */
    private $fields = [];
    /**
     * @var array параметр json-api запроса sort
     */
    private $sort = [];
    /**
     * @var array|mixed параметр json-api запроса filter
     */
    private $filter = [];
    /**
     * @var array|mixed параметр json-api запроса page
     */
    private $page = [];

    /** @var object экземпляр сервиса у которого будет вызван метода при get() */
    private $restService;

    /** @var string название метода севриса, который будет вызван при вызове get() */
    private $serviceMethod;

    /**
     * RestQuery constructor.
     * @param Request|null $request запрос из которого взять данные
     */
    public function __construct(?Request $request = null)
    {
        if ($request) {
            $rawIncludes = $request->get('include');
            $this->includes = $rawIncludes ? explode(',', $rawIncludes): [];

            $rawFields = $request->get('fields', []);

            foreach ($rawFields as $code => $values) {
                $this->fields[$code] = explode(',', $values);
            }


            $this->filter = $request->get('filter', []);

            $rawSort = $request->get('sort');
            $this->sort = $rawSort ? explode(',', $rawSort) : [];

            $this->page = $request->get('page', []);
        }
    }

    // construct query

    /**
     * Задать пагинацию в режиме страниц
     * @param int $number
     * @param int $size
     * @return RestQuery
     */
    public function pageNumber(int $number, int $size): self
    {
        $this->page = [
            'number' => $number,
            'size' => $size,
        ];

        return $this;
    }

    /**
     * Задать пагинацию в режиме смещения
     * @param int $offset
     * @param int $limit
     * @return RestQuery
     */
    public function pageOffset(int $offset, int $limit): self
    {
        $this->page = [
            'offset' => $offset,
            'limit' => $limit,
        ];

        return $this;
    }

    /**
     * Указать выводимые поля для конкретной сущнсоти
     * @param string $code
     * @param mixed ...$fields
     * @return RestQuery
     */
    public function addFields(string $code, ...$fields): self
    {
        $this->fields[$code] = array_merge($this->fields[$code] ?? [], $fields);

        return $this;
    }

    /**
     * Добавить relations
     * @param mixed ...$codes - символьный код связи
     * @return RestQuery
     */
    public function include(...$codes): self
    {
        $this->includes = array_merge($this->includes, $codes);

        return $this;
    }

    /**
     * Задать фильтр по полю
     * Пример:
     *      setFilter('id', 123)
     *      setFilter('name', 'like', 'John')
     * @param string $field название поля
     * @param mixed $op операция или значение (тогда операция равна '=')
     * @param mixed|null $value значение
     * @return RestQuery
     */
    public function setFilter(string $field, $op, $value = null): self
    {
        if ($value == null) {
            if (in_array($op, ['null', 'notNull'])) {
                $value = '';
            } else {
                $value = $op;
                $op = '=';
            }
        }
        $this->filter[$this->packFilterField($field, $op)] = $value;

        return $this;
    }

    /**
     * Добавить сортировку по полю
     * @param string $field
     * @param string $direction (asc or desc)
     * @return RestQuery
     */
    public function addSort(string $field, string $direction = 'asc'): self
    {
        $this->sort[] = ($direction == 'desc' ? '-' : '') . $field;

        return $this;
    }

    // read query

    /**
     * Проверить требуется ли вывод связанной сущности
     * @param string $code
     * @return bool
     */
    public function isIncluded(string $code): bool
    {
        return in_array($code, $this->includes);
    }

    /**
     * Получить список запрашиваемых полей для типа ресурса
     * @param string $code
     * @return array
     */
    public function getFields(string $code): array
    {
        return $this->fields[$code] ?? [];
    }

    /**
     * Получить итератор для обхода всех условий фильтрации.
     * Каждое условие представляет собой массив из трёх элементов [$field, $op, $value]
     * Пример:
     *      foreach ($q->filterIterator() as [$field, $op, $value]) {
     *          //...
     *      }
     * @return Generator
     */
    public function filterIterator(): Generator
    {
        foreach ($this->filter as $code => $value) {
            $result = $this->unpackFilterField($code);
            $result[] = $value;
            yield $result;
        }
    }

    /**
     * Получить список условий для заданного поля.
     * @param string $fieldName
     * @return array
     */
    public function getFilter(string $fieldName): array
    {
        $result = [];
        foreach ($this->filterIterator() as [$field, $op, $value]) {
            if ($fieldName == $field) {
                $result[] = [$op, $value];
            }
        }

        return $result;
    }

    /**
     * Удалить все условия для заданного поля.
     * @param string $name
     */
    public function removeFilter(string $name): void
    {
        foreach ($this->filter as $key => $value) {
            [$field, $op] = $this->unpackFilterField($key);
            if ($field == $name) {
                unset($this->filter[$key]);
            }
        }
    }

    /**
     * Удалить условия для выводимых полей для конкретной сущнсоти.
     * @param string $name
     */
    public function removeField(string $name): void
    {
        unset($this->fields[$name]);
    }

    /**
     * Получить настройки пагинации.
     * Независимо от способа передачи параметров пагинации (number, offset) всегда возвразает
     * результат в формате [offset, limit]
     * @return array|null
     */
    public function getPage(): ?array
    {
        if (isset($this->page['number']) && isset($this->page['size'])) {
            return [
                'offset' => ($this->page['number'] - 1) * $this->page['size'],
                'limit' => $this->page['size'],
            ];
        } elseif (isset($this->page['offset']) && isset($this->page['limit'])) {
            return $this->page;
        } else {
            return null;
        }
    }

    /**
     * Получить итератор для обхода списка сортировки.
     * Элемент сортировки представляет из мебя массив, где на первом месте всегда название поля, а на втором символьный код направелния (asc, desc)
     * @return Generator
     */
    public function sortIterator(): Generator
    {
        foreach ($this->sort as $code) {
            $desc = preg_match("/^-.*/", $code);
            yield [
                $desc ? substr($code, 1) : $code,
                $desc ? 'desc' : 'asc',
            ];
        }
    }

    /**
     * @param string $fieldName
     * @return array
     */
    public function getSort(string $fieldName): array
    {
        $result = [];
        foreach ($this->sortIterator() as [$field, $direction]) {
            if ($fieldName == $field) {
                $result[] = [$direction];
            }
        }

        return $result;
    }

    /**
     * @param string $name
     */
    public function removeSort(string $name): void
    {
        foreach ($this->sort as $key => $code) {
            if ($code == $name || $code == "-{$name}") {
                unset($this->sort[$key]);
            }
        }
    }

    // misc

    /**
     * Сформировать код фильтра из имени и операции
     * @param  string  $field
     * @param  string  $op
     * @return string
     */
    protected function packFilterField(string $field, string $op): string
    {
        return $op == '=' ? $field : $field . '|' . $op;
    }

    /**
     * Разбить код фильтра на имя и операцию
     * @param  string  $code
     * @return array
     */
    protected function unpackFilterField(string $code): array
    {
        $data = explode('|', $code);
        $field = $data[0];
        $op = count($data) == 1 ? '=' : $data[1];

        return [$field, $op];
    }

    /**
     * Настроить запрос для дальнейшего вызова get()
     * @param object $service экземпляр рест-сервисаа
     * @return RestQuery
     */
    public function prepare($service)
    {
        $this->restService = $service;

        return $this;
    }

    /**
     * Завершить конструирование запроса и выполнить его.
     * @param string $name название метода rest-клиента
     * @param array $args дополнительные агрументы метода rest-клиента
     * @return mixed
     * @throws \Exception если в RestQuery не задан rest-клиент
     */
    public function __call($name, $args)
    {
        if (!$this->restService) {
            throw new \Exception("RestQuery не сконфигурирован для вызова __call()");
        }

        return $this->restService->{$name}($this, ...$args);
    }

    /**
     * Сформировать массив, пригодный для передачи его в качестве http query params
     * @return array
     */
    public function toQuery(): array
    {
        $query = [];
        if (count($this->includes)) {
            $query['include'] = join(',', $this->includes);
        }
        if (count($this->filter)) {
            $query['filter'] = $this->filter;
        }
        if (count($this->fields)) {
            $query['fields'] = [];
            foreach ($this->fields as $code => $fields) {
                $query['fields'][$code] = join(',', $fields);
            }
        }
        if (count($this->sort)) {
            $query['sort'] = join(',', $this->sort);
        }
        if (count($this->page)) {
            $query['page'] = $this->page;
        }

        return $query;
    }
}
