<?php

namespace App\Core\Rest;

use Closure;
use Psr\Http\Message\ResponseInterface;

class RestResult
{
    /** @var ResponseInterface */
    private $response;
    /** @var array|Closure[] */
    private $expectedStatuses;
    /** @var Closure */
    private $unexpectedStatusAction;
    /** @var string|null */
    private $dataPath;
    /** @var Closure|string */
    private $wrapper;
    /** @var bool */
    private $wrapCollection;

    public function __construct(ResponseInterface $response)
    {
        $this->response = $response;
    }

    public function expectStatus(int $code, ?Closure $action = null): self
    {
        $this->expectedStatuses[$code] = $action;

        return $this;
    }

    public function unexpectedStatus(Closure $action): self
    {
        $this->unexpectedStatusAction = $action;

        return $this;
    }

    public function wrapResult($wrapper, bool $collection = false, ?string $path = null): self
    {
        $this->wrapper = $wrapper;
        $this->dataPath = $path;
        $this->wrapCollection = $collection;

        return $this;
    }

    /**
     * @param string $exceptionClass
     * @return mixed
     */
    public function value(?string $exceptionClass = null)
    {
        $value = $this->checkStatuses($exceptionClass);
        if ($value) {
            return $value;
        }
        $data = data_get(json_decode($this->response->getBody()->getContents(), true), $this->dataPath);

        if (!$this->wrapper) {
            return $data;
        }

        if ($this->wrapCollection) {
            $result = collect();
            foreach ($data as $item) {
                $result[] = $this->wrapDataPiece($item, $this->wrapper);
            }

            return $result;
        } else {
            return $this->wrapDataPiece($data, $this->wrapper);
        }
    }

    /**
     * @param string $exceptionClass
     * @return mixed
     */
    public function checkStatuses(?string $exceptionClass = null)
    {
        $statusCode = $this->response->getStatusCode();
        if (!array_key_exists($statusCode, $this->expectedStatuses)) {
            if (!$this->unexpectedStatusAction) {
                if (!$exceptionClass) {
                    $exceptionClass = \Exception::class;
                }

                throw new $exceptionClass('Unexpected status code without action');
            }

            return call_user_func($this->unexpectedStatusAction, $this->response);
        }

        $action = $this->expectedStatuses[$statusCode];
        if ($action) {
            return call_user_func($action, $this->response);
        }

        return null;
    }

    private function wrapDataPiece($data, $wrapper)
    {
        if ($wrapper instanceof Closure) {
            return call_user_func($wrapper, $data);
        } elseif ($wrapper) {
            return new $wrapper($data);
        } else {
            return $data;
        }
    }
}
