<?php

namespace App\Domain\Categories\Models;

use App\Domain\Support\Models\Model;
use Cviebrock\EloquentSluggable\Sluggable;
use Database\Factories\Categories\CategoriesFactory;
use Illuminate\Database\Eloquent\Builder;
use Illuminate\Database\Eloquent\Relations\BelongsTo;
use Illuminate\Database\Eloquent\Relations\BelongsToMany;
use Illuminate\Database\Eloquent\Relations\HasMany;
use Illuminate\Support\Collection;
use Kalnoy\Nestedset\NodeTrait;

/**
 * Класс-модель для сущности "Категории товара".
 *
 * @property string $name - название
 * @property int $parent_id - id родительской категории
 * @property string|null $file - файл иконки категории
 * @property string $icon - код категории иконки для витрины
 * @property string $code - ЧПУ код категории
 *
 * @property-read Category|null $parent - родительская категория
 * @property-read Category|null $parentCategory - родительская категория (алиас для parent)
 * @property-read Collection|Category[]|null $ancestors - все родительские категории
 *
 * @property-read Collection|Category[]|null $children - дочерние категории
 * @property-read Collection|Category[]|null $categories - дочерние категории (алиас для children)
 * @property-read Collection|Category[]|null $descendants - все дочерние категории
 *
 * @property-read Collection|Property[] $properties - свойства (характеристики) товаров
 *
 * @method static Collection|Category[] ancestorsAndSelf(int $id)
 * @method static Builder|self whereAncestorOf(int $categoryId, bool $self)
 */
class Category extends Model
{
    use NodeTrait;
    use Sluggable;

    public const FILLABLE = ['name', 'code', 'parent_id'];

    protected $table = 'categories';
    protected $fillable = self::FILLABLE;

    /**
     * @return BelongsTo
     */
    public function parentCategory(): BelongsTo
    {
        return $this->parent();
    }

    /**
     * @return HasMany
     */
    public function categories(): HasMany
    {
        return $this->children();
    }

    /**
     * @return BelongsToMany
     */
    public function properties(): BelongsToMany
    {
        return $this->belongsToMany(Property::class, (new CategoryPropertyLink())->getTable(), 'category_id', 'property_id');
    }

    public function allProperties()
    {
        $properties = $this->properties;
        $parent = $this->parent;
        while ($parent) {
            $properties = $properties->merge($parent->properties);
            $parent = $parent->parent;
        }

        return $properties;
    }

    /**
     * Return the sluggable configuration array for this model.
     *
     * @return array
     */
    public function sluggable(): array
    {
        return [
            'code' => [
                'source' => 'name',
                'separator' => '_',
                'unique' => true,
                'maxLength' => 30,
            ],
        ];
    }

    public function replicate(array $except = null): self
    {
        $defaults = [
            $this->getParentIdName(),
            $this->getLftName(),
            $this->getRgtName(),
            'code',
        ];

        $except = empty($except) ? $defaults : array_unique(array_merge($except, $defaults));
        return parent::replicate($except);
    }

    protected static function boot()
    {
        parent::boot();

        self::deleted(function (self $category) {
            Category::query()->fixTree();
        });
    }

    public static function factory(): CategoriesFactory
    {
        return CategoriesFactory::new();
    }
}
