<?php

namespace App\Domain\Categories\Models;

use App\Domain\Support\Models\Model;
use Illuminate\Database\Eloquent\Builder;
use Illuminate\Database\Eloquent\Collection;
use Illuminate\Database\Eloquent\Relations\BelongsTo;
use Illuminate\Support\Facades\DB;

/**
 * Класс-модель для сущности "Привязки свойств (характеристик) к категориям товаров"
 * Class CategoryPropertyLink
 * @package App\Models
 *
 * @property int $property_id - id свойства
 * @property int $category_id - id категории
 * @property bool $is_required - обязательность
 *
 * @property-read Property $property - свойство
 * @property-read Category $category - категория
 *
 * @method static Builder|self forCategories(array $categoryIds)
 */
class CategoryPropertyLink extends Model
{
    const TABLE_NAME = 'category_property_links';

    /**
     * @var array Заполняемые поля модели
     */
    const FILLABLE = ['property_id', 'category_id', 'is_required'];

    protected $fillable = self::FILLABLE;

    protected $table = self::TABLE_NAME;

    protected static $restIncludes = ['property', 'category'];

    public static function getByCategoryIds(array $categoryIds): Collection
    {
        return self::forCategories($categoryIds)
            ->select('id', 'property_id', 'category_id')
            ->get();
    }

    /**
     * @return BelongsTo
     */
    public function property(): BelongsTo
    {
        return $this->belongsTo(Property::class);
    }

    /**
     * @return BelongsTo
     */
    public function category(): BelongsTo
    {
        return $this->belongsTo(Category::class);
    }

    public function scopeForCategories(Builder $query, array $categoryIds): Builder
    {
        return $query->whereIn('category_id', $categoryIds);
    }

    /**
     * Возвращает массив идентификаторов привязок действующих для категории свойств.
     *
     * @param int $categoryId
     * @return array|int[]
     */
    public static function selectEffectiveIds(int $categoryId): array
    {
        return Category::whereAncestorOf($categoryId, true)
            ->join(self::TABLE_NAME.' as link', 'categories.id', '=', 'link.category_id')
            ->select([
                'link.id as link_id',
                'link.property_id as property_id',
                DB::raw('CASE WHEN link.is_required THEN 0 else categories._rgt END as priority')
            ])
            ->orderBy('priority')
            ->get()
            ->unique('property_id')
            ->pluck('link_id')
            ->toArray();
    }

    public static function make(Category|int $category, Property|int $property, bool $isRequired = false): static
    {
        $instance = new static();
        $instance->category_id = is_int($category) ? $category : $category->id;
        $instance->property_id = is_int($property) ? $property : $property->id;
        $instance->is_required = $isRequired;

        return $instance;
    }
}
