<?php

namespace App\Domain\Products\Actions\Attributes;

use App\Domain\Categories\Models\CategoryPropertyView;
use App\Domain\Categories\Models\Property;
use App\Domain\Products\Models\Product;
use App\Domain\Products\Models\ProductPropertyValue;
use App\Domain\Support\Concerns\InteractsWithModels;
use Illuminate\Support\Collection;

class SetProductAttributesAction
{
    use InteractsWithModels;

    /**
     * @param Product $product
     * @param array $values атрибуты в виде [property_id => value(s)]
     * @param bool $clearUnassigned если True, значения непереданных в $values атрибутов будут удалены
     */
    public function execute(Product $product, array $values, bool $clearUnassigned): void
    {
        $attributes = $this->loadAvailableAttributes($product);
        $existingValues = $this->loadExistingValues($product);

        foreach ($attributes as $attribute) {
            if (!array_key_exists($attribute->id, $values)) {
                continue;
            }

            $this->setValues(
                $product,
                $attribute,
                $existingValues->pull($attribute->id) ?? new Collection(),
                $this->extractNewValues($attribute, $values)
            );
        }

        if ($clearUnassigned) {
            $existingValues->each(fn(Collection $oldValues) => $this->deleteValues($oldValues));
        }
    }

    protected function loadAvailableAttributes(Product $product): Collection
    {
        return CategoryPropertyView::forCategory($product->category_id)->get();
    }

    protected function loadExistingValues(Product $product): Collection
    {
        return $product->properties()
            ->get()
            ->groupBy('property_id');
    }

    protected function extractNewValues(Property $property, array $values): array
    {
        $propertyValues = array_filter((array)$values[$property->id]);

        if (!empty($propertyValues) && !$property->is_multiple) {
            $propertyValues = array_slice($propertyValues, 0, 1);
        }

        return $propertyValues;
    }

    protected function setValues(Product $product, Property $attribute, Collection $oldValues, array $newValues): void
    {
        foreach ($newValues as $newValue) {
            $value = $oldValues->isEmpty() ? ProductPropertyValue::make($product, $attribute) : $oldValues->shift();
            $value->setValue($attribute->type, $newValue);

            $this->saveOrThrow($value);
        }

        $this->deleteValues($oldValues);
    }

    protected function deleteValues(Collection $values): void
    {
        $values->each(fn(ProductPropertyValue $value) => $value->delete());
    }
}
