<?php

namespace App\Domain\Products\Models;

use App\Domain\Categories\Models\Category;
use App\Domain\Categories\Models\Property;
use App\Domain\Classifiers\Models\Brand;
use App\Domain\Classifiers\Models\Country;
use App\Domain\Classifiers\Models\Manufacturer;
use App\Domain\Classifiers\Models\ProductType;
use App\Domain\Support\Models\Model;
use App\Models\Offer;
use App\Models\ProductConstraint;
use App\Models\ProductTip;
use Carbon\Carbon;
use Cviebrock\EloquentSluggable\Sluggable;
use Database\Factories\Products\ProductsFactory;
use Illuminate\Database\Eloquent\Builder;
use Illuminate\Database\Eloquent\Relations\BelongsTo;
use Illuminate\Database\Eloquent\Relations\HasMany;
use Illuminate\Support\Collection;

/**
 * Класс-модель для сущности "Товары"
 * Class Product
 * @package App\Models
 *
 * @property string $external_id - id товара в системе
 * @property int $brand_id - id бренда
 * @property int $manufacturer_id - id производителя
 * @property int $country_id - id страны
 * @property int $category_id - id категории
 * @property int $product_type_id - id типа товара ("весовой", "штучный", ...)
 * @property string $barcode - артикул (EAN)
 * @property string $name - название
 * @property string $code - ЧПУ код
 * @property string $description - описание
 * @property float $weight - вес нетто (кг)
 * @property float $weight_gross - вес брутто (кг)
 * @property float $width - длина (мм)
 * @property float $height - высота (мм)
 * @property float $length - ширина (мм)
 * @property ProductArchiveType $archive - тип архивности
 * @property string $archive_comment - комментарий для архивного товара
 * @property Carbon $archive_date - дата архивации
 * @property bool $is_new - признак новинки
 * @property string $ingredients - состав товара (в свободной форме)
 * @property int $calories - калорийность на 100 г, ккал
 * @property float $proteins - белки на 100 г, г
 * @property float $fats - жиры на 100 г, г
 * @property float $carbohydrates - углеводы на 100 г, г
 * @property string $storage_area - условия хранения
 * @property string $shelf_life - срок хранения
 * @property string $additional_info - дополнительная информация
 * @property int $vat_rate - ставка НДС, в процентах
 *
 * @property int $variant_group_id
 *
 * @property-read Category $category - категория
 * @property-read Brand $brand - бренд
 * @property-read Manufacturer $manufacturer - производитель
 * @property-read ProductType $productType - тип товара
 * @property-read Country $country - страна производства
 * @property-read Collection|ProductPropertyValue[] $properties - свойства
 * @property-read Collection|Offer[] $offers - офферы продукта
 * @property-read Collection|ProductTip[] $tips - особенности товара
 * @property-read VariantGroup $variantGroup - группа вариантов товара
 * @property-read Collection|ProductImage[] $images - изображения товара
 * @property-read Collection|ProductConstraint[] $constraints - ограничения по доставке и оплате
 *
 * @method static Builder|self inCategories(array|int $categoryIds)
 * @method static Builder|self inVariantGroup(int $groupId)
 */
class Product extends Model
{
    use Sluggable;

    /** Заполняемые поля модели */
    const FILLABLE = [
        'is_new', 'brand_id', 'category_id', 'barcode', 'name', 'code', 'description',
        'weight', 'width', 'height', 'length', 'archive', 'archive_comment', 'archive_date',
        'ingredients', 'calories', 'proteins', 'fats', 'carbohydrates', 'storage_area',
        'shelf_life', 'additional_info', 'vat_rate', 'manufacturer_id', 'country_id',
        'weight_gross', 'product_type_id', 'external_id',
    ];

    /**
     * @var array
     */
    protected $fillable = self::FILLABLE;

    /** @var bool */
    protected static $unguarded = true;

    /** @var array */
    protected $casts = [
        'weight' => 'float',
        'weight_gross' => 'float',
        'proteins' => 'float',
        'fats' => 'float',
        'carbohydrates' => 'float',
        'width' => 'float',
        'height' => 'float',
        'length' => 'float',
        'archive' => ProductArchiveType::class,
    ];

    /** @var array */
    protected $dates = ['archive_date'];

    /**
     * @param int $id
     * @param array $fields
     * @return static|null
     */
    public static function getOneActive(int $id, array $fields = ['*']): ?self
    {
        return self::query()
            ->where('id', $id)
            ->where('archive', ProductArchiveType::NO_ARCHIVE)
            ->first($fields);
    }

    /**
     * @return bool
     */
    public function isActiveForShowcase(): bool
    {
        return !$this->archive->isArchive();
    }

    /**
     * @return array
     */
    public function sluggable(): array
    {
        return [
            'code' => [
                'source' => 'name',
            ],
        ];
    }

    /**
     * @return BelongsTo
     */
    public function category(): BelongsTo
    {
        return $this->belongsTo(Category::class);
    }

    /**
     * Устанавливает связь с брендом.
     *
     * @return BelongsTo
     */
    public function brand(): BelongsTo
    {
        return $this->belongsTo(Brand::class);
    }

    /**
     * Устанавливает связь с производителем товара.
     *
     * @return BelongsTo
     */
    public function manufacturer(): BelongsTo
    {
        return $this->belongsTo(Manufacturer::class);
    }

    /**
     * Устанавливает связь со страной производства товара.
     *
     * @return BelongsTo
     */
    public function country(): BelongsTo
    {
        return $this->belongsTo(Country::class);
    }

    /**
     * Устанавливает связь с типом товара.
     *
     * @return BelongsTo
     */
    public function productType(): BelongsTo
    {
        return $this->belongsTo(ProductType::class);
    }

    /**
     * @return HasMany
     */
    public function properties(): HasMany
    {
        return $this->hasMany(ProductPropertyValue::class, 'product_id');
    }

    /**
     * @return HasMany
     */
    public function offers(): HasMany
    {
        return $this->hasMany(Offer::class, 'product_id');
    }

    /**
     * @return HasMany
     */
    public function tips(): HasMany
    {
        return $this->hasMany(ProductTip::class);
    }

    /**
     * @return BelongsTo
     */
    public function variantGroup(): BelongsTo
    {
        return $this->belongsTo(VariantGroup::class);
    }

    /**
     * @return HasMany
     */
    public function images(): HasMany
    {
        return $this->hasMany(ProductImage::class);
    }

    /**
     * @return HasMany
     */
    public function constraints(): HasMany
    {
        return $this->hasMany(ProductConstraint::class);
    }

    /**
     * @param callable $callback - function (Property $property, ?ProductPropertyValue $value)
     * @param Property[] $properties
     * @return array
     */
    public function variantInfo($callback, array $properties)
    {
        $result = [];
        $propertyValuesMap = $this->properties->keyBy('property_id');
        foreach ($properties as $property) {
            $value = $propertyValuesMap[$property->id] ?? null;
            $result[] = call_user_func($callback, $property, $value);
        }

        return $result;
    }

    public function variantKey(array $properties)
    {
        $keys = $this->variantInfo(function (Property $property, ?ProductPropertyValue $propertyValue) {
            $value = $propertyValue ? $propertyValue->value($property->type) : '!@#$%';

            return $property->id . '#' . $value;
        }, $properties);
        sort($keys);

        return join('|', $keys);
    }

    public function variantValues(Collection $attributes, string $placeholder = '!@#$%'): string
    {
        $result = [];
        $allValues = $this->properties->groupBy('property_id');

        foreach ($attributes as $attribute) {
            $attributeValues = $allValues->get($attribute->id);
            if ($attributeValues === null) {
                $result[] = $placeholder;

                continue;
            }

            $values = [];
            foreach ($attributeValues as $attributeValue) {
                $values[] = (string)($attributeValue->value($attribute->type) ?? $placeholder);
            }
            sort($values);
            $result = array_merge($result, $values);
        }

        return implode('|', $result);
    }

    public static function factory(): ProductsFactory
    {
        return ProductsFactory::new();
    }

    public function scopeInCategories(Builder $query, mixed $categoryIds): Builder
    {
        $categoryIds = array_filter(
            array_map('intval', (array)$categoryIds)
        );
        $categoryWithChildrenIds = Category::query()
            ->select('id', '_lft', '_rgt')
            ->with('descendants')
            ->whereIn('id', $categoryIds)
            ->get()
            ->flatMap(fn (Category $category) => $category->descendants->pluck('id')->push($category->id))
            ->unique()
            ->toArray();

        return empty($categoryIds) ? $query : $query->whereIn('category_id', $categoryWithChildrenIds);
    }

    public function scopeInVariantGroup(Builder $query, mixed $groupId): Builder
    {
        return $query->where('variant_group_id', $groupId);
    }
}
