<?php

namespace App\Domain\Products\Models;

use App\Domain\Categories\Models\Property;
use App\Domain\Categories\Models\PropertyDirectoryValue;
use App\Domain\Categories\Models\PropertyType;
use App\Domain\Support\Models\Model;
use Illuminate\Database\Eloquent\Relations\BelongsTo;
use Illuminate\Support\Carbon;

/**
 * Класс-модель для сущности "Значения свойств (характеристик) товаров"
 * Class ProductPropertyValue
 * @package App\Models
 *
 * @property int $property_id - id свойства
 * @property int $product_id - id товара
 * @property PropertyType $type - тип значения
 * @property string|null $value_string - значение для строкового свойства
 * @property int|null $value_integer - значение для целочисленного свойства
 * @property float|null $value_double - значение для вещественного свойства
 * @property Carbon|null $value_datetime - значение для свойства даты и времени
 * @property int|null $value_directory - значение для свойства-справочника
 * @property bool|null $value_boolean - логическое значение
 * @property string|null $value_link - значение типа ссылка
 *
 * @property-read Property $property - свойство
 * @property-read Product $product - товар
 * @property-read PropertyDirectoryValue $propertyDirectoryValue - вариант значения свойства-справочника
 */
class ProductPropertyValue extends Model
{
    /**
     * @var array Заполняемые поля модели
     */
    const FILLABLE = [
        'property_id',
        'product_id',
        'type',
        'value_string',
        'value_integer',
        'value_double',
        'value_datetime',
        'value_directory',
        'value_link',
        'value_boolean',
    ];

    protected $fillable = self::FILLABLE;

    protected $table = 'product_property_values';

    protected $dates = ['value_datetime'];

    protected $casts = [
        'value_int' => 'int',
        'value_directory' => 'int',
        'value_double' => 'float',
        'type' => PropertyType::class,
    ];

    public static function typeToField(PropertyType $propertyType): string
    {
        return "value_{$propertyType->value}";
    }

    public static function make(Product|int $product, Property $property, mixed $value = null): self
    {
        $instance = new static();
        $instance->product_id = is_int($product) ? $product : $product->id;
        $instance->property_id = $property->id;

        if ($value !== null) {
            $instance->setValue($property->type, $value);
        }

        return $instance;
    }

    public function value(?PropertyType $type = null): mixed
    {
        return $this->{self::typeToField($type ?? $this->type)};
    }

    public function setValue(PropertyType $type, $value): void
    {
        $this->{self::typeToField($type)} = $value;
        $this->type = $type;
    }

    public function getValueLinkAttribute(): ?string
    {
        return $this->attributes['value_string'] ?? null;
    }

    public function setValueLinkAttribute(?string $value): void
    {
        $this->attributes['value_string'] = $value;
        $this->type = PropertyType::LINK();
    }

    public function getValueBooleanAttribute(): ?bool
    {
        if (!isset($this->attributes['value_integer'])) {
            return null;
        }
        return $this->attributes['value_integer'] == 1;
    }

    public function setValueBooleanAttribute(?bool $value): void
    {
        $this->type = PropertyType::BOOLEAN();
        if ($value === null) {
            $this->attributes['value_integer'] = null;
            return;
        }

        $this->attributes['value_integer'] = $value == true ? 1 : 0;
    }

    /**
     * @return BelongsTo
     */
    public function property(): BelongsTo
    {
        return $this->belongsTo(Property::class);
    }

    /**
     * @return BelongsTo
     */
    public function product(): BelongsTo
    {
        return $this->belongsTo(Product::class);
    }

    /**
     * @return BelongsTo
     */
    public function propertyDirectoryValue(): BelongsTo
    {
        return $this->belongsTo(PropertyDirectoryValue::class, 'value_directory');
    }
}
