<?php

namespace App\Http\ApiV1\Modules\Products\Controllers;

use App\Domain\Products\Actions\Products\CreateProductAction;
use App\Domain\Products\Actions\Products\DeleteManyProductsAction;
use App\Domain\Products\Actions\Products\DeleteProductAction;
use App\Domain\Products\Actions\Products\MarkProductsForIndexAction;
use App\Domain\Products\Actions\Products\PatchProductAction;
use App\Domain\Products\Actions\Attributes\PatchProductAttributesAction;
use App\Domain\Products\Actions\Products\ReplaceProductAction;
use App\Domain\Products\Actions\Attributes\ReplaceProductAttributesAction;
use App\Http\ApiV1\Modules\Products\Queries\FilterProductsQuery;
use App\Http\ApiV1\Modules\Products\Queries\ProductsQuery;
use App\Http\ApiV1\Modules\Products\Requests\CreateOrReplaceProductRequest;
use App\Http\ApiV1\Modules\Products\Requests\PatchProductRequest;
use App\Http\ApiV1\Modules\Products\Requests\ReplaceProductAttributesRequest;
use App\Http\ApiV1\Modules\Products\Resources\CountResource;
use App\Http\ApiV1\Modules\Products\Resources\ProductsResource;
use App\Http\ApiV1\Support\Pagination\PageBuilderFactory;
use App\Http\ApiV1\Support\Requests\MassDeleteRequest;
use App\Http\ApiV1\Support\Resources\EmptyResource;
use Illuminate\Http\Request;
use Illuminate\Http\Resources\Json\AnonymousResourceCollection;

class ProductsController
{
    public function search(PageBuilderFactory $pageBuilderFactory, ProductsQuery $query): AnonymousResourceCollection
    {
        return ProductsResource::collectPage(
            $pageBuilderFactory->fromQuery($query)->build()
        );
    }

    public function get(int $entityId, ProductsQuery $query): ProductsResource
    {
        return new ProductsResource($query->findOrFail($entityId));
    }

    public function create(CreateOrReplaceProductRequest $request, CreateProductAction $action): ProductsResource
    {
        return new ProductsResource($action->execute($request->validated()));
    }

    public function replace(
        int $entityId,
        CreateOrReplaceProductRequest $request,
        ReplaceProductAction $action
    ): ProductsResource {
        return new ProductsResource($action->execute($entityId, $request->validated()));
    }

    public function patch(int $entityId, PatchProductRequest $request, PatchProductAction $action): ProductsResource
    {
        return new ProductsResource($action->execute($entityId, $request->validated()));
    }

    public function delete(int $entityId, DeleteProductAction $action): EmptyResource
    {
        $action->execute($entityId);
        return new EmptyResource();
    }

    public function massDelete(MassDeleteRequest $request, DeleteManyProductsAction $action): EmptyResource
    {
        $action->execute($request->getIds());
        return new EmptyResource();
    }

    public function replaceAttributes(
        int $entityId,
        ReplaceProductAttributesRequest $request,
        ReplaceProductAttributesAction $action
    ): ProductsResource {
        return new ProductsResource(
            $action->enrichResult()->execute($entityId, $request->values())
        );
    }

    public function patchAttributes(
        int $entityId,
        ReplaceProductAttributesRequest $request,
        PatchProductAttributesAction $action
    ): ProductsResource {
        return new ProductsResource(
            $action->enrichResult()->execute($entityId, $request->values())
        );
    }

    public function markForIndex(FilterProductsQuery $query, MarkProductsForIndexAction $action): CountResource
    {
        // Выполняем валидацию запроса и заодно определяем наличие товаров
        $count = $query->count();

        if ($count > 0) {
            $action->onQueue()->execute($query->serialize());
        }

        return new CountResource($count);
    }
}
