<?php

namespace App\Http\ApiV1\Modules\Products\Controllers;

use App\Core\Elastic\Category\CategoryIndex;
use App\Core\Elastic\IndexFactory;
use App\Core\Elastic\Product\ProductArrayExtractor;
use App\Core\Elastic\Product\ProductIndex;
use App\Core\Elastic\Product\ProductQuery;
use App\Core\Elastic\ScheduleRunner;
use function blank;
use Elasticsearch\Client;
use Illuminate\Http\JsonResponse;
use Illuminate\Http\Request;
use function resolve;
use function response;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;

/**
 * Контроллер для работы с поиском (Эластиком).
 *
 * Class SearchController
 * @package App\Http\Controllers\V1
 */
class SearchController
{
    /**
     * Запрашивает категории у Эластика.
     *
     * @param Request $request
     * @return JsonResponse
     * @deprecated
     */
    public function categories(Request $request)
    {
        $query = $request->get('query');
        $limit = $request->get('limit', 10);
        if (!$query) {
            throw new BadRequestHttpException('query is required');
        }
        $factory = new IndexFactory(resolve(Client::class));
        /** @var CategoryIndex $index */
        $index = $factory->index(IndexFactory::TYPE_CATEGORY);
        $ids = $index->search($query, $limit);

        return response()->json($ids);
    }

    /**
     * Запрашивает продукты у Эластика.
     *
     * Параметры запроса:
     * page, offset, limit, id, seller_id, offer_id, active, saleActive, priceFrom, priceTo,
     * qtyFrom, qtyTo, code, name, categoryCode, brandCode, fullText, archive, props, barcode,
     * dateFrom, dateTo, orderBy, fields, aggregations.
     *
     * Неиспользуемые параметры, от которых в перспективе надо избавиться:
     * groupCode, isNew.
     *
     * TODO надо бы отрефакторить и разбить на подметоды.
     *
     * @param Request $request
     * @return JsonResponse
     */
    public function products(Request $request)
    {
        $factory = new IndexFactory(resolve(Client::class));
        /** @var ProductIndex $index */
        $index = $factory->index(IndexFactory::TYPE_PRODUCT);

        $productQuery = new ProductQuery();
        if ($page = $request->get('page')) {
            $productQuery->page($page, $request->get('pageSize'));
        }

        $offset = $request->get('offset');
        $limit = $request->get('limit');
        if ($offset !== null && $limit !== null) {
            $productQuery->pageOffset($offset, $limit);
        }

        $id = $request->get('id');
        if ($id) {
            $productQuery->id($id);
        }

        $seller_id = $request->get('seller_id');
        if ($seller_id) {
            $productQuery->seller_id($seller_id);
        }

        $offer_id = $request->get('offer_id');
        if ($offer_id) {
            $productQuery->offer_id($offer_id);
        }

        $isActive = $request->get('active');
        if (!blank($isActive)) {
            if ($seller_id || $offer_id) {
                $productQuery->active($isActive);
            } else {
                $productQuery->productActive($isActive);
                $productQuery->hasActiveOffers($isActive);
            }
        }

        $isSaleActive = $request->get('saleActive');
        if (!blank($isSaleActive)) {
            $productQuery->saleActive($isSaleActive);
        }

        $priceFrom = $request->get('priceFrom');
        $priceTo = $request->get('priceTo');
        if ($priceFrom || $priceTo) {
            $productQuery->price($priceFrom, $priceTo);
        }

        $qtyFrom = $request->get('qtyFrom');
        $qtyTo = $request->get('qtyTo');
        if ($qtyFrom || $qtyTo) {
            $productQuery->qty($qtyFrom, $qtyTo);
        }

        $code = $request->get('code');
        if ($code) {
            $productQuery->code($code);
        }

        $name = $request->get('name');
        if ($name) {
            $productQuery->name($name);
        }

        $categoryCode = $request->get('categoryCode');
        if ($categoryCode) {
            $productQuery->category($categoryCode);
        }

        $brandCodes = $request->get('brandCode');
        if ($brandCodes) {
            $productQuery->brand($brandCodes);
        }

        $groupCode = $request->get('groupCode');
        if ($groupCode) {
            $productQuery->productGroup($groupCode);
        }

        $isNew = $request->get('isNew');
        if ($isNew) {
            $productQuery->new();
        }

        $fullText = $request->get('fullText');
        if ($fullText) {
            $productQuery->fullText($fullText);
        }

        $props = $request->get('props');
        if ($props) {
            foreach ($props as $name => $values) {
                $productQuery->property($name, $values);
            }
        }

        $archive = $request->get('archive');
        if (!blank($archive)) {
            $productQuery->archive($archive);
        }

        $barcode = $request->get('barcode');
        if ($barcode) {
            $productQuery->barcode($barcode);
        }

        $dateFrom = $request->get('dateFrom');
        $dateTo = $request->get('dateTo');
        if ($dateFrom || $dateTo) {
            $productQuery->created($dateFrom, $dateTo);
        }

        $order = $request->get('orderBy');
        if ($order && is_array($order)) {
            foreach ($order as $field => $direction) {
                switch ($field) {
                    case 'price':
                        $productQuery->orderByPrice($seller_id, $offer_id, $direction);

                        break;
                    case 'rating':
                        $productQuery->orderBy('rating', $direction);

                        break;
                    case 'dateAdd':
                        $productQuery->orderBy('date_add', $direction);

                        break;
                }
            }
        }

        $fields = $request->get('fields', []);
        $indexMark = false;
        if (($key = array_search('indexMark', $fields)) !== false) {
            $indexMark = true;
            unset($fields[$key]);
        }

        $aggregations = $request->get('aggregations', []);
        if (($key = array_search('categoryCounts', $aggregations)) !== false) {
            $productQuery->aggregateCategoryCounts();
        }

        $searchResult = $index->search($productQuery, new ProductArrayExtractor($fields, $seller_id, $offer_id));
        if ($indexMark) {
            $ids = [];
            foreach ($searchResult['items'] as $item) {
                $ids[] = $item['id'];
            }
            $scheduler = new ScheduleRunner();
            $notReadyIds = $scheduler->checkMarked(IndexFactory::TYPE_PRODUCT, $ids);

            foreach ($searchResult['items'] as $i => $item) {
                $item['ready'] = !in_array($item['id'], $notReadyIds);
                $searchResult['items'][$i] = $item;
            }
        }

        return response()->json($searchResult);
    }
}
