<?php

namespace App\Models;

use App\Core\Rest\Eloquent\AbstractRestModel;
use App\Core\Rest\RestQuery;
use App\Domain\Products\Models\Product;
use Illuminate\Database\Eloquent\Builder;
use Illuminate\Database\Eloquent\Collection;
use Illuminate\Database\Eloquent\Relations\BelongsTo;
use Illuminate\Database\Eloquent\Relations\HasMany;

/**
 * Класс-модель для сущности "Предложения мерчантов".
 *
 * Class Offer
 * @package App\Models
 * @property int $product_id - id товара
 * @property int $seller_id - id продавца
 * @property string $external_id - id товара у мерчанта
 * @property int $sale_status - статус продажи
 * @property string $storage_address - адрес хранения товара в магазине (для сборщика)
 *
 * @property-read Product $product - товар
 * @property-read Collection|OfferCert[] $certs
 */
class Offer extends AbstractRestModel
{
    /**
     * Заполняемые поля модели
     */
    const FILLABLE = ['product_id', 'seller_id', 'external_id', 'sale_status', 'storage_address'];

    /**
     * @var array
     */
    protected $fillable = self::FILLABLE;

    /**
     * @var array
     */
    protected static $restIncludes = ['product', 'certs'];

    /**
     * @return BelongsTo
     */
    public function product(): BelongsTo
    {
        return $this->belongsTo(Product::class);
    }

    /**
     * @param Builder $query
     * @param RestQuery $restQuery
     * @return Builder
     */
    public static function modifyQuery(Builder $query, RestQuery $restQuery): Builder
    {
        $productNameFilter = $restQuery->getFilter('productName');
        $product = [];
        if ($productNameFilter) {
            [$op, $value] = $productNameFilter[0];
            $product['name'] = $value;
            $restQuery->removeFilter('productName');
        }

        $productExternalIdFilter = $restQuery->getFilter('productExternalId');
        if ($productExternalIdFilter) {
            [$op, $value] = $productExternalIdFilter[0];
            $product['external_id'] = $value;
            $restQuery->removeFilter('productExternalId');
        }

        if ($product) {
            $query->whereHas('product', function (Builder $query) use ($product) {
                foreach ($product as $key => $value) {
                    switch ($key) {
                        case 'name':
                            $query->where($key, 'like', "%{$value}%");

                            break;
                        default:
                            $query->where($key, $value);
                    }
                }
            });
        }

        return parent::modifyQuery($query, $restQuery);
    }

    /**
     * Получить активные офферы товаров.
     * @param int[] $productIds
     * @return array|[productId => [sellerId => Offer]]
     */
    public static function getActiveByProducts(array $productIds): array
    {
        /** @var Collection[] $activeOfferGroups */
        $activeOfferGroups = self::query()
            ->whereIn('product_id', $productIds)
            ->get();
        $result = [];
        /** @var Offer $offer */
        foreach ($activeOfferGroups as $offer) {
            $result[$offer->product_id][$offer->seller_id] = $offer;
        }

        return $result;
    }

    /**
     * @return HasMany
     * @deprecated
     */
    public function certs(): HasMany
    {
        return $this->hasMany(OfferCert::class, 'offer_id');
    }
}
