<?php

namespace App\Providers;

use App\Core\Elastic\Product\Indexing\ProductsPipeline;
use App\Core\RequestInitiator\HeaderUserIdProvider;
use App\Core\RequestInitiator\UserIdProvider;
use App\Http\Middleware\RecognizeSubject;
use App\Support\Extensions\OptionalMacros;
use App\Support\Files\FileStorage;
use App\Support\Log\DefaultLoggerAwareInterface;
use Carbon\CarbonImmutable;
use Elasticsearch\Client;
use Elasticsearch\ClientBuilder;
use Ensi\LaravelAuditing\Resolvers\SubjectManager;
use Ensi\LaravelEnsiFilesystem\EnsiFilesystemManager;
use Illuminate\Contracts\Foundation\Application;
use Illuminate\Database\Eloquent\Model;
use Illuminate\Support\DateFactory;
use Illuminate\Support\Facades\Log;
use Illuminate\Support\Optional;
use Illuminate\Support\ServiceProvider;
use ReflectionException;
use Storage;

class AppServiceProvider extends ServiceProvider
{
    /**
     * Register any application services.
     *
     * @return void
     */
    public function register()
    {
        DateFactory::useClass(CarbonImmutable::class);

        $this->app->singleton(RecognizeSubject::class, function (Application $app) {
            return new RecognizeSubject(
                $app->make(SubjectManager::class),
                $app->make('config')->get('ensi-audit.subjects')
            );
        });

        $this->app->singleton(FileStorage::class, function (Application $app) {
            return new FileStorage(Storage::disk($app->make(EnsiFilesystemManager::class)->publicDiskName()));
        });

        $this->app->singleton(ProductsPipeline::class);

        $this->app->afterResolving(
            DefaultLoggerAwareInterface::class,
            fn (DefaultLoggerAwareInterface $resolved) => $resolved->setLogger(Log::channel())
        );
    }

    /**
     * Bootstrap any application services.
     *
     * @return void
     * @throws ReflectionException
     */
    public function boot(): void
    {
        Model::preventLazyLoading(!app()->isProduction());
        Optional::mixin(new OptionalMacros());

        $this->app->singleton(Client::class, function () {
            return ClientBuilder::create()
                ->setHosts([config('elasticsearch.host')])
                ->build();
        });

        $this->app->singleton(UserIdProvider::class, HeaderUserIdProvider::class);
    }
}
