<?php

namespace Tests;

use Illuminate\Database\Eloquent\Model;
use Illuminate\Foundation\Testing\DatabaseTransactions;
use Illuminate\Testing\TestResponse;

abstract class APITestCase extends TestCase
{
    use DatabaseTransactions;

    /**
     * @param string $route
     * @param Model|int|null $resource
     * @return TestResponse
     */
    protected function apiGet(string $route, $resource): TestResponse
    {
        return $this->json('GET', $this->makeResourceApiUrl($route, $resource));
    }

    /**
     * @param string $route
     * @param array $data
     * @param Model|int|null $parent
     * @return TestResponse
     */
    protected function apiCreate(string $route, array $data, $parent = null): TestResponse
    {
        $url = $parent === null
            ? route($route, [], false)
            : $this->makeResourceApiUrl($route, $parent);

        return $this->json('POST', $url, $data);
    }

    /**
     * @param string $route
     * @param Model|int|null $resource
     * @param array $data
     * @return TestResponse
     */
    protected function apiReplace(string $route, $resource, array $data): TestResponse
    {
        return $this->json('PUT', $this->makeResourceApiUrl($route, $resource), $data);
    }

    /**
     * @param string $route
     * @param Model|int|null $resource
     * @param array $data
     * @return TestResponse
     */
    protected function apiPatch(string $route, $resource, array $data): TestResponse
    {
        return $this->json('PATCH', $this->makeResourceApiUrl($route, $resource), $data);
    }

    /**
     * @param string $route
     * @param Model|int|null $resource
     * @return TestResponse
     */
    protected function apiDelete(string $route, $resource): TestResponse
    {
        return $this->json('DELETE', $this->makeResourceApiUrl($route, $resource));
    }

    protected function apiSearch(string $route, array $filter = [], array $include = []): TestResponse
    {
        $data = [];

        if (!empty($filter)) {
            $data['filter'] = $filter;
        }

        if (!empty($include)) {
            $data['include'] = $include;
        }

        return $this->json('POST', route($route, [], false), $data);
    }

    /**
     * @param string $route
     * @param array|int[] $ids
     * @return TestResponse
     */
    protected function apiMassDelete(string $route, array $ids): TestResponse
    {
        return $this->json('POST', route($route, [], false), ['ids' => $ids]);
    }

    /**
     * @param string $route
     * @param Model|int|null $resource
     * @param Model|int|null $nested
     * @return string
     */
    protected function makeResourceApiUrl(string $route, $resource, $nested = null): string
    {
        $params = [
            'entityId' => $this->getApiResourceId($resource),
            'resourceId' => $this->getApiResourceId($nested),
        ];

        return route($route, array_filter($params), false);
    }

    /**
     * @param Model|int|null $resource
     * @return int|null
     */
    protected function getApiResourceId($resource): ?int
    {
        if ($resource === null) {
            return null;
        }

        return $resource instanceof Model
            ? $resource->getKey()
            : (int)$resource;
    }
}
