<?php

namespace Tests\Feature\Categories;

use App\Domain\Categories\Models\Category;
use App\Domain\Categories\Models\CategoryPropertyLink;
use App\Domain\Categories\Models\Property;
use App\Domain\Categories\Models\PropertyDirectoryValue;
use App\Domain\Categories\Models\PropertyType;
use Tests\APITestCase;

class CategoriesTest extends APITestCase
{
    private static $jsonStructure = ['id', 'code', 'name', 'parent_id'];

    public function testCreate(): void
    {
        $data = Category::factory()->makeFillableData();

        $response = $this->apiCreate('createCategory', $data);

        $response->assertStatus(201);
        $response->assertJsonStructure(['data' => self::$jsonStructure]);
        $response->assertJsonFragment($data);
    }

    public function testCreateSubCategory(): void
    {
        $parent = Category::factory()->createOne();
        $assignParent = ['parent_id' => $parent->id];
        $data = Category::factory()->makeFillableData($assignParent);

        $response = $this->apiCreate('createCategory', $data);

        $response->assertStatus(201);
        $response->assertJsonFragment($assignParent);
    }

    public function testReplace(): void
    {
        $category = Category::factory()->createOne();
        $data = Category::factory()->makeFillableData();

        $response = $this->apiReplace('replaceCategory', $category, $data);

        $response->assertStatus(200);
        $response->assertJsonStructure(['data' => self::$jsonStructure]);
        $response->assertJsonFragment($data);
    }

    public function testPatch(): void
    {
        $category = Category::factory()->createOne();
        $data = ['name' => 'New patch name'];

        $response = $this->apiPatch('patchCategory', $category, $data);

        $response->assertStatus(200);
        $response->assertJsonStructure(['data' => self::$jsonStructure]);
        $response->assertJsonFragment($data);
    }

    public function testPatchParent(): void
    {
        $parents = Category::factory()->count(2)->create();
        $category = Category::factory()->withParent($parents[0])->createOne();
        $data = ['parent_id' => $parents[1]->id];

        $response = $this->apiPatch('patchCategory', $category, $data);

        $response->assertStatus(200);
        $response->assertJsonStructure(['data' => self::$jsonStructure]);
        $response->assertJsonFragment($data);
    }

    public function testDelete(): void
    {
        $category = Category::factory()->createOne();

        $response = $this->apiDelete('deleteCategory', $category);

        $response->assertStatus(200);
        $response->assertJsonFragment(['data' => null]);
        $this->assertNull(Category::find($category->id));
    }

    public function testMassDelete(): void
    {
        $ids = Category::factory()
            ->count(3)
            ->create()
            ->pluck('id')
            ->toArray();

        $response = $this->apiMassDelete('deleteCategories', $ids);

        $response->assertStatus(200);
        $response->assertJsonFragment(['data' => null]);

        $this->assertCount(0, Category::find($ids));
    }

    public function testGet(): void
    {
        $category = Category::factory()->createOne();

        $response = $this->apiGet('getCategory', $category);

        $response->assertStatus(200);
        $response->assertJsonStructure(['data' => self::$jsonStructure]);
    }

    public function testSearch(): void
    {
        $category = Category::factory()->count(3)->create()[1];

        $response = $this->apiSearch('searchCategories', ['code' => $category->code]);

        $response->assertStatus(200);
        $response->assertJsonCount(1, 'data');
        $response->assertJsonFragment(['name' => $category->name]);
    }

    public function testProperties(): void
    {
        $category = Category::factory()->createOne();
        $property = Property::factory()->createOne(['type' => PropertyType::DIRECTORY]);
        $value = PropertyDirectoryValue::factory()->forProperty($property)->createOne();

        CategoryPropertyLink::create([
            'category_id' => $category->getKey(),
            'property_id' => $property->getKey(),
            'is_required' => true,
        ]);

        $query = [
            'include' => ['directory'],
        ];
        $response = $this->json('POST', $this->makeResourceApiUrl('getCategoryProperties', $category), $query);

        $response->assertStatus(200);
        $response->assertJsonCount(1, 'data');
        $response->assertJsonFragment([
            'id' => $property->id,
            'name' => $property->name,
            'is_required' => true,
            'is_inherited' => false,
            'directory' => [[
                'id' => $value->id,
                'code' => $value->code,
                'name' => $value->name,
            ]],
        ]);
    }

    public function testTreeWithRoot(): void
    {
        Category::factory()->createOne();
        $root = Category::factory()->createTree(3)->first();

        $query = [
            'filter' => ['root_id' => $root->id]
        ];

        $response = $this->json('POST', route('getCategoriesTree', [], false), $query);

        $response->assertStatus(200);
        $response->assertJsonStructure([
            'data' => [['id', 'children' => [['id', 'children' => [['id', 'children']]]]]]
        ]);
        $response->assertJsonFragment(['id' => $root->id, 'name' => $root->name]);
        $this->assertCount(1, $response->json('data'));
    }
}
