<?php

namespace Tests\Feature\Categories;

use App\Domain\Categories\Models\Property;
use App\Domain\Categories\Models\PropertyDirectoryValue;
use App\Domain\Categories\Models\PropertyType;
use Tests\APITestCase;

class PropertiesTest extends APITestCase
{
    private static $jsonStructure = [
        'id',
        'code',
        'name',
        'display_name',
        'type',
        'is_multiple',
        'is_filterable',
        'is_color',
    ];

    public function testCreate(): void
    {
        $data = Property::factory()->makeFillableData();

        $response = $this->apiCreate('createProperty', $data);

        $response->assertStatus(201);
        $response->assertJsonStructure(['data' => self::$jsonStructure]);
        $response->assertJsonFragment($data);
    }

    public function testReplace(): void
    {
        $property = Property::factory()->createOne();
        $data = Property::factory()->makeFillableData();

        $response = $this->apiReplace('replaceProperty', $property, $data);

        $response->assertStatus(200);
        $response->assertJsonStructure(['data' => self::$jsonStructure]);
        $response->assertJsonFragment($data);
    }

    public function testPatch(): void
    {
        $property = Property::factory()->createOne();
        $data = ['name' => 'New patch name'];

        $response = $this->apiPatch('patchProperty', $property, $data);

        $response->assertStatus(200);
        $response->assertJsonStructure(['data' => self::$jsonStructure]);
        $response->assertJsonFragment($data);
    }

    public function testDelete(): void
    {
        $property = Property::factory()->createOne();

        $response = $this->apiDelete('deleteProperty', $property);

        $response->assertStatus(200);
        $response->assertJsonFragment(['data' => null]);
        $this->assertNull(Property::find($property->id));
    }

    public function testMassDelete(): void
    {
        $ids = Property::factory()
            ->count(3)
            ->create()
            ->pluck('id')
            ->toArray();

        $response = $this->apiMassDelete('deleteProperties', $ids);

        $response->assertStatus(200);
        $response->assertJsonFragment(['data' => null]);

        $this->assertCount(0, Property::find($ids));
    }

    public function testGet(): void
    {
        $property = Property::factory()->createOne();

        $response = $this->apiGet('getProperty', $property);

        $response->assertStatus(200);
        $response->assertJsonStructure(['data' => self::$jsonStructure]);
    }

    public function testSearch(): void
    {
        $property = Property::factory()->count(3)->create()[1];

        $response = $this->apiSearch('searchProperties', ['code' => $property->code]);

        $response->assertStatus(200);
        $response->assertJsonCount(1, 'data');
        $response->assertJsonFragment(['name' => $property->name]);
    }

    public function testSearchIncludesDirectory(): void
    {
        $property = Property::factory()->createOne(['type' => PropertyType::DIRECTORY]);
        PropertyDirectoryValue::factory()->forProperty($property)->createOne();

        $expectedJsonStruct = array_merge(
            self::$jsonStructure,
            ['directory' => [['id', 'code', 'name']]]
        );

        $response = $this->apiSearch('searchProperties', ['code' => $property->code], ['directory']);

        $response->assertStatus(200);
        $response->assertJsonStructure(['data' => [$expectedJsonStruct]]);
    }
}
