<?php

namespace Tests\Feature;

use App\Domain\Categories\Models\Category;
use App\Domain\Classifiers\Models\Brand;
use App\Domain\Products\Models\Product;
use App\Domain\Products\Models\ProductImage;
use App\Support\Files\FileStorage;
use Database\Factories\Factory;
use Ensi\LaravelEnsiFilesystem\Models\EnsiFile;
use Illuminate\Http\UploadedFile;
use Illuminate\Testing\TestResponse;
use Mockery;
use Tests\APITestCase;

class FilesTest extends APITestCase
{
    private string $fakePath;

    protected function setUp(): void
    {
        parent::setUp();

        $this->fakePath = "path/to/test1.png";

        $mockStorage = Mockery::mock(FileStorage::class);
        $mockStorage->shouldReceive('add')->andReturn($this->fakePath);
        $mockStorage->shouldReceive('remove');

        $this->app->instance(FileStorage::class, $mockStorage);
    }

    /**
     * @dataProvider provideFiles
     */
    public function testUpload(string $name, Factory $factory, string $field): void
    {
        $this->assertUpload("upload$name", $factory->createOne(), $field);
    }

    /**
     * @dataProvider provideFiles
     */
    public function testDelete(string $name, Factory $factory, string $field): void
    {
        $this->assertDelete("delete$name", $factory->createOne([$field => $this->fakePath]));
    }

    public function provideFiles(): array
    {
        return [
            'Brand image' => ['BrandImage', Brand::factory(), 'file'],
            'Category image' => ['CategoryImage', Category::factory(), 'file'],
        ];
    }

    public function testCreateProductImage(): void
    {
        $product = Product::factory()->createOne();
        $image = ProductImage::factory()->owned($product)->createOne();

        $response = $this->assertUpload('uploadProductImage', $image);

        $response->assertJsonStructure(['data' => ['id', 'type', 'file']]);
    }

    private function assertUpload(
        string $route,
        $resource,
        string $field = 'file'
    ): TestResponse {
        $data = [$field => UploadedFile::fake()->image('test.png')];
        $response = $this->json(
            'POST',
            $this->makeResourceApiUrl($route, $resource),
            $data
        );

        $response->assertStatus(200);
        $response->assertJsonFragment(['file' => EnsiFile::public($this->fakePath)->jsonSerialize()]);

        return $response;
    }

    private function assertDelete(string $route, $resource): void
    {
        $response = $this->json(
            'POST',
            $this->makeResourceApiUrl($route, $resource),
            []
        );

        $response->assertStatus(200);
        $response->assertJsonFragment(['file' => null]);
    }
}
