<?php

namespace Tests\Feature\Products;

use App\Domain\Categories\Models\Category;
use App\Domain\Categories\Models\CategoryPropertyLink;
use App\Domain\Categories\Models\Property;
use App\Domain\Categories\Models\PropertyDirectoryValue;
use App\Domain\Categories\Models\PropertyType;
use App\Domain\Classifiers\Models\Brand;
use App\Domain\Classifiers\Models\ProductType;
use App\Domain\Products\Actions\Products\MarkProductsForIndexAction;
use App\Domain\Products\Models\Product;
use App\Domain\Products\Models\ProductPropertyValue;
use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\Queue;
use Spatie\QueueableAction\Testing\QueueableActionFake;
use Tests\APITestCase;

class ProductsTest extends APITestCase
{
    private static $jsonStructure = [
        'id',
        'code',
        'name',
        'description',
        'product_type_id',
        'brand_id',
        'weight',
        'archive',
    ];

    public function testCreate(): void
    {
        $data = Product::factory()->makeFillableData();

        $response = $this->apiCreate('createProduct', $data);

        $response->assertStatus(201);
        $response->assertJsonStructure(['data' => self::$jsonStructure]);
        $response->assertJsonFragment($data);
    }

    public function testReplace(): void
    {
        $product = Product::factory()->createOne();
        $data = Product::factory()->makeFillableData();

        $response = $this->apiReplace('replaceProduct', $product, $data);

        $response->assertStatus(200);
        $response->assertJsonStructure(['data' => self::$jsonStructure]);
        $response->assertJsonFragment($data);
    }

    public function testPatch(): void
    {
        $product = Product::factory()->createOne();
        $data = ['name' => 'New patch name'];

        $response = $this->apiPatch('patchProduct', $product, $data);

        $response->assertStatus(200);
        $response->assertJsonStructure(['data' => self::$jsonStructure]);
        $response->assertJsonFragment($data);
    }

    public function testDelete(): void
    {
        $product = Product::factory()->createOne();

        $response = $this->apiDelete('deleteProduct', $product);

        $response->assertStatus(200);
        $response->assertJsonFragment(['data' => null]);
        $this->assertNull(Product::find($product->id));
    }

    public function testMassDelete(): void
    {
        $ids = Product::factory()
            ->count(3)
            ->create()
            ->pluck('id')
            ->toArray();

        $response = $this->apiMassDelete('deleteProducts', $ids);

        $response->assertStatus(200);
        $response->assertJsonFragment(['data' => null]);

        $this->assertCount(0, Product::find($ids));
    }

    public function testGet(): void
    {
        $product = Product::factory()->createOne();

        $response = $this->apiGet('getProduct', $product);

        $response->assertStatus(200);
        $response->assertJsonStructure(['data' => self::$jsonStructure]);
    }

    public function testSearch(): void
    {
        $product = Product::factory()->count(3)->create()[1];

        $response = $this->apiSearch('searchProducts', ['code' => $product->code]);

        $response->assertStatus(200);
        $response->assertJsonCount(1, 'data');
        $response->assertJsonFragment(['name' => $product->name]);
    }

    public function testSearchInSubCategories(): void
    {
        $root = Category::factory()->createOne();
        $parent = Category::factory()->withParent($root)->createOne();
        $category = Category::factory()->withParent($parent)->createOne();
        $product = Product::factory()->createOne(['category_id' => $category->id]);

        $response = $this->apiSearch('searchProducts', ['category_id' => $root->id]);

        $response->assertStatus(200);
        $response->assertJsonCount(1, 'data');
        $response->assertJsonFragment(['id' => $product->id]);
    }

    public function testSearchIncludes(): void
    {
        $brand = Brand::factory()->createOne();
        $type = ProductType::factory()->createOne();

        $product = Product::factory()->createOne([
            'brand_id' => $brand->getKey(),
            'product_type_id' => $type->getKey(),
        ]);

        $expectedJson = array_merge(self::$jsonStructure, [
            'brand' => ['id', 'code', 'name'],
            'type' => ['id', 'code', 'name'],
        ]);

        $response = $this->apiSearch(
            'searchProducts',
            ['id' => [$product->id]],
            ['brand', 'type']
        );

        $response->assertStatus(200);
        $response->assertJsonStructure(['data' => [$expectedJson]]);
    }

    public function testSearchIncludeAttributes(): void
    {
        $property = Property::factory()->createOne(['is_multiple' => true, 'type' => PropertyType::INTEGER]);
        $product = Product::factory()->createOne();
        ProductPropertyValue::make($product, $property, 25)->save();
        ProductPropertyValue::make($product, $property, 100)->save();

        $expectedJson = [
            'attributes' => [
                ['property_id' => $property->id, 'value' => [25, 100], 'type' => PropertyType::INTEGER]
            ]
        ];

        $response = $this->apiSearch(
            'searchProducts',
            ['id' => [$product->id]],
            ['attributes']
        );

        $response->assertStatus(200);
        $response->assertJsonFragment($expectedJson);
    }

    public function testSearchIncludeAttributeDirectoryValues(): void
    {
        $property = Property::factory()->createOne(['type' => PropertyType::DIRECTORY]);
        $directory = PropertyDirectoryValue::factory()->forProperty($property)->createOne();
        $product = Product::factory()->createOne();
        ProductPropertyValue::make($product, $property, $directory->id)->save();

        $expectedJson = [
            'directory' => [
                ['id' => $directory->id, 'code' => $directory->code, 'name' => $directory->name],
            ]
        ];

        $response = $this->apiSearch(
            'searchProducts',
            ['id' => [$product->id]],
            ['attributes']
        );

        $response->assertStatus(200);
        $response->assertJsonFragment($expectedJson);
    }

    public function testReplaceAttributes(): void
    {
        $category = Category::factory()->createOne();
        $property = Property::factory()->createOne(['type' => PropertyType::INTEGER]);
        $product = Product::factory()->createOne(['category_id' => $category->id]);
        CategoryPropertyLink::make($category, $property)->save();

        $data = [
            'values' => [['property_id' => $property->id, 'value' => [2]]]
        ];

        $response = $this->json('PUT', $this->makeResourceApiUrl('replaceProductAttributes', $product), $data);

        $response->assertStatus(200);
        $response->assertJsonStructure([
            'data' => array_merge(self::$jsonStructure, ['attributes' => [['property_id']]])
        ]);
    }

    public function testMarkForIndex(): void
    {
        Queue::fake();
        $product = Product::factory()->createOne();

        $response = $this->apiSearch('markForIndexProducts', ['code' => $product->code]);

        $response->assertStatus(200);
        $response->assertJsonFragment(['count' => 1]);
        QueueableActionFake::assertPushed(MarkProductsForIndexAction::class);
    }

    public function testMarkForIndexNotFoundsProducts(): void
    {
        Queue::fake();

        $response = $this->apiSearch('markForIndexProducts', ['code' => 'unknown']);

        $response->assertStatus(200);
        $response->assertJsonFragment(['count' => 0]);
        QueueableActionFake::assertNotPushed(MarkProductsForIndexAction::class);
    }

    public function testMarkForIndexInvalid(): void
    {
        Queue::fake();
        $product = Product::factory()->createOne();

        $params = [
            'code' => $product->code,
            'unknown' => 12,
        ];
        $response = $this->apiSearch('markForIndexProducts', $params);

        $response->assertStatus(400);
        QueueableActionFake::assertNotPushed(MarkProductsForIndexAction::class);
    }
}
