<?php

namespace Tests\Feature\Products;

use App\Domain\Categories\Models\Property;
use App\Domain\Categories\Models\PropertyType;
use App\Domain\Products\Models\Product;
use App\Domain\Products\Models\ProductPropertyValue;
use App\Domain\Products\Models\VariantGroup;
use Tests\APITestCase;

class VariantGroupsTest extends APITestCase
{
    private static $jsonStructure = ['id', 'product_ids', 'attribute_ids'];

    public function testCreate(): void
    {
        $attribute = Property::factory()->createOne(['type' => PropertyType::INTEGER]);
        $product = Product::factory()->createOne();
        ProductPropertyValue::make($product, $attribute, 100)->save();

        $data = ['product_ids' => [$product->id], 'attribute_ids' => [$attribute->id]];

        $response = $this->apiCreate('createVariantGroup', $data);

        $response->assertStatus(201);
        $response->assertJsonStructure(['data' => self::$jsonStructure]);
        $response->assertJsonFragment($data);
    }

    public function testReplace(): void
    {
        $group = $this->createVariantGroup();
        $attribute = Property::factory()->createOne();

        $data = ['product_ids' => $group->product_ids, 'attribute_ids' => [$attribute->id]];

        $response = $this->apiReplace('replaceVariantGroup', $group, $data);

        $response->assertStatus(200);
        $response->assertJsonStructure(['data' => self::$jsonStructure]);
        $response->assertJsonFragment($data);
    }

    public function testPatch(): void
    {
        $group = $this->createVariantGroup();
        $attribute = Property::factory()->createOne();

        $data = ['attribute_ids' => [$attribute->id]];

        $response = $this->apiPatch('replaceVariantGroup', $group, $data);

        $response->assertStatus(200);
        $response->assertJsonStructure(['data' => self::$jsonStructure]);
        $response->assertJsonFragment($data);
    }

    public function testDelete(): void
    {
        $group = $this->createVariantGroup();

        $response = $this->apiDelete('deleteVariantGroup', $group);

        $response->assertStatus(200);
        $response->assertJsonFragment(['data' => null]);
    }

    public function testGet(): void
    {
        $group = $this->createVariantGroup();

        $response = $this->apiGet('getVariantGroup', $group);

        $response->assertStatus(200);
        $response->assertJsonStructure(['data' => self::$jsonStructure]);
    }

    public function testSearch(): void
    {
        $group = $this->createVariantGroup();

        $expectedJsonStruct = array_merge(self::$jsonStructure, [
            'attributes' => [['code', 'type', 'id']],
            'products' => [['name', 'attributes' => [['value']]]],
        ]);

        $response = $this->apiSearch(
            'searchVariantGroups',
            ['id' => $group->id],
            ['products.attributes', 'attributes']
        );

        $response->assertStatus(200);
        $response->assertJsonStructure(['data' => [$expectedJsonStruct]]);
    }

    private function createVariantGroup(): VariantGroup
    {
        $attribute = Property::factory()->createOne(['type' => PropertyType::INTEGER]);
        $product = Product::factory()->createOne();
        ProductPropertyValue::make($product, $attribute, 128)->save();

        $group = VariantGroup::create([
            'product_ids' => [$product->id],
            'attribute_ids' => [$attribute->id],
        ]);

        $product->update(['variant_group_id' => $group->id]);
        $group->properties()->sync($group->attribute_ids);

        return $group;
    }
}
