<?php

namespace Tests\Integration\Api\Categories;

use App\Domain\Categories\Models\Category;
use App\Domain\Categories\Models\CategoryPropertyLink;
use App\Domain\Categories\Models\CategoryPropertyView;
use App\Domain\Categories\Models\Property;
use App\Domain\Categories\Models\PropertyDirectoryValue;
use App\Domain\Categories\Models\PropertyType;
use App\Http\ApiV1\Modules\Categories\Queries\CategoryPropertiesQueryFactory;
use Illuminate\Foundation\Testing\DatabaseTransactions;
use Illuminate\Http\Request;
use Spatie\QueryBuilder\QueryBuilder;
use Tests\TestCase;

class CategoryPropertiesQueryFactoryTest extends TestCase
{
    use DatabaseTransactions;

    public function testCreateWithoutParams(): void
    {
        $category = Category::factory()->createOne();
        $this->createProperty($category);

        $results = $this->createQuery($category, [])->get();

        $this->assertCount(1, $results);
    }

    public function testExcludeInherited(): void
    {
        $parent = Category::factory()->createOne();
        $child = Category::factory()->withParent($parent)->createOne();

        $this->createProperty($parent, true);
        $this->createProperty($child, true);

        $data = [
            'filter' => ['is_inherited' => false],
        ];
        $results = $this->createQuery($child, $data)->get();

        $this->assertCount(1, $results);
        $this->assertFalse($results->first()->is_inherited);
    }

    public function testSelectsOnlyRequired(): void
    {
        $category = Category::factory()->createOne();
        $this->createProperty($category);
        $this->createProperty($category, true);

        $data = [
            'filter' => ['is_required' => true],
        ];

        $results = $this->createQuery($category, $data)->get();

        $this->assertCount(1, $results);
        $this->assertTrue($results->first()->is_required);
    }

    public function testSelectsFilterable(): void
    {
        $category = Category::factory()->createOne();
        $expected = $this->createProperty($category, true, ['is_filterable' => true]);
        $this->createProperty($category, true, ['is_filterable' => false]);

        $data = [
            'filter' => ['is_filterable' => true],
        ];

        $results = $this->createQuery($category, $data)->get();

        $this->assertEquals([$expected->getKey()], $results->pluck('id')->toArray());
    }

    public function testSelectsByType(): void
    {
        $category = Category::factory()->createOne();
        $expected = $this->createProperty($category, true, ['type' => PropertyType::BOOLEAN()]);
        $this->createProperty($category, true, ['type' => PropertyType::DATETIME()]);

        $data = [
            'filter' => ['type' => PropertyType::BOOLEAN],
        ];

        $results = $this->createQuery($category, $data)->get();

        $this->assertEquals([$expected->getKey()], $results->pluck('id')->toArray());
    }

    public function testIncludesDirectory(): void
    {
        $category = Category::factory()->createOne();
        $property = $this->createProperty($category, true, ['type' => PropertyType::DIRECTORY()]);

        $directory = PropertyDirectoryValue::factory()->forProperty($property)->count(2)->create();
        $data = [
            'include' => ['directory'],
        ];

        /** @var CategoryPropertyView $result */
        $result = $this->createQuery($category, $data)->get()->first();

        $this->assertTrue($result->relationLoaded('directory'));
        $this->assertEquals($directory->count(), $result->directory->count());
    }

    public function testIgnoresSiblings(): void
    {
        $parent = Category::factory()->createOne();
        $child = Category::factory()->withParent($parent)->createOne();
        $sibling = Category::factory()->withParent($parent)->createOne();

        $parentProp = $this->createProperty($parent);
        $childProp = $this->createProperty($child);
        $this->createProperty($sibling);

        $results = $this->createQuery($child)->get();

        $this->assertEqualsCanonicalizing(
            [$parentProp->id, $childProp->id],
            $results->map->id->toArray()
        );
    }

    public function testMissingProperties(): void
    {
        $parent = Category::factory()->createOne();
        $child = Category::factory()->withParent($parent)->createOne();
        $sibling = Category::factory()->withParent($parent)->createOne();
        $this->createProperty($sibling);

        $results = $this->createQuery($child)->get();

        $this->assertCount(0, $results);
    }

    private function createQuery(Category $category, array $requestData = []): QueryBuilder
    {
        $request = new Request($requestData);
        $factory = new CategoryPropertiesQueryFactory();

        return $factory->createQuery($category->getKey(), $request);
    }

    private function createProperty(?Category $linkTo = null, bool $isRequired = false, array $fields = []): Property
    {
        $property = Property::factory()->createOne($fields);

        if ($linkTo != null) {
            $this->createLink($property, $linkTo, $isRequired);
        }

        return $property;
    }

    private function createLink(Property $property, Category $linkTo, bool $isRequired = false): CategoryPropertyLink
    {
        return CategoryPropertyLink::create([
            'category_id' => $linkTo->getKey(),
            'property_id' => $property->getKey(),
            'is_required' => $isRequired,
        ]);
    }
}
