<?php

namespace Tests\Integration\Domain\Products\Actions;

use App\Domain\Categories\Models\Category;
use App\Domain\Categories\Models\CategoryPropertyLink;
use App\Domain\Categories\Models\Property;
use App\Domain\Categories\Models\PropertyType;
use App\Domain\Products\Actions\Attributes\SetProductAttributesAction;
use App\Domain\Products\Models\Product;
use App\Domain\Products\Models\ProductPropertyValue;
use Illuminate\Foundation\Testing\DatabaseTransactions;
use Illuminate\Support\Collection;
use Illuminate\Support\Facades\Date;
use Tests\TestCase;

class SetProductAttributesActionTest extends TestCase
{
    use DatabaseTransactions;

    /** @var Category */
    private $category;

    /** @var Product */
    private $product;

    /** @var SetProductAttributesAction */
    private $testing;

    protected function setUp(): void
    {
        parent::setUp();

        $this->category = Category::factory()->createOne();
        $this->product = Product::factory()->createOne(['category_id' => $this->category->id]);

        $this->testing = new SetProductAttributesAction();
    }

    /**
     * @dataProvider provideAttributesValues
     */
    public function testSetValues(PropertyType $type, bool $isMultiple, mixed $values, array $expected): void
    {
        $property = $this->createLinkedAttribute($type, $isMultiple);

        $this->executeAction([$property->id => $values]);

        $this->assertPropertyValuesEquals($expected, $property);
    }

    public function provideAttributesValues(): array
    {
        return [
            'single value' => [PropertyType::INTEGER(), false, 125, [125]],
            'multiple values' => [PropertyType::INTEGER(), true, [125, 50], [125, 50]],
            'single with multiple values' => [PropertyType::INTEGER(), false, [125, 50], [125]],
            'null value multiple property' => [PropertyType::INTEGER(), true, null, []],
            'null value single property' => [PropertyType::INTEGER(), false, null, []],
            'datetime value' => [
                PropertyType::DATETIME(),
                false,
                '2021-05-15T14:48:11.000000Z',
                [Date::make('2021-05-15 14:48:11')]
            ],
        ];
    }

    public function testReplacesOldValues(): void
    {
        $property = $this->createLinkedAttribute(PropertyType::INTEGER(), true);
        $this->createAttributesValues($property, [219, 11, 556]);

        $this->executeAction([$property->id => [145, 11]]);

        $this->assertPropertyValuesEquals([145, 11], $property);
    }

    public function testClearsOldValues(): void
    {
        $property = $this->createLinkedAttribute(PropertyType::INTEGER(), true);
        $this->createAttributesValues($property, [219]);

        $this->executeAction([$property->id => []]);

        $this->assertPropertyValuesEquals([], $property);
    }

    public function testClearsUnassigned(): void
    {
        $property = $this->createLinkedAttribute(PropertyType::STRING());
        $this->createAttributesValues($property, ['Value1']);

        $this->executeAction([], true);

        $this->assertPropertyValuesEquals([], $property);
    }

    public function testNotClearsUnassigned(): void
    {
        $property = $this->createLinkedAttribute(PropertyType::STRING());
        $this->createAttributesValues($property, ['Value1']);

        $this->executeAction([]);

        $this->assertPropertyValuesEquals(['Value1'], $property);
    }

    private function createLinkedAttribute(PropertyType $type, bool $isMultiple = false): Property
    {
        $property = Property::factory()->createOne([
            'is_multiple' => $isMultiple,
            'type' => $type,
        ]);

        CategoryPropertyLink::create([
            'property_id' => $property->id,
            'category_id' => $this->category->id,
            'is_required' => false,
        ]);

        return $property;
    }

    private function createAttributesValues(Property $property, array $values): Collection
    {
        return Collection::make($values)
            ->map(fn(mixed $value) => ProductPropertyValue::make($this->product, $property, $value))
            ->each(fn(ProductPropertyValue $value) => $value->save());
    }

    private function executeAction(array $values, bool $replace = false): void
    {
        $this->testing->execute($this->product, $values, $replace);
    }

    private function assertPropertyValuesEquals(array $expected, Property $property): void
    {
        $actual = $this->product->properties()
            ->where('property_id', $property->id)
            ->get()
            ->map(fn(ProductPropertyValue $value) => $value->value($property->type))
            ->toArray();

        $this->assertEqualsCanonicalizing($expected, $actual);
    }
}
