<?php

namespace Tests\Integration\Domain\Products\Actions;

use App\Domain\Categories\Models\Property;
use App\Domain\Categories\Models\PropertyType;
use App\Domain\Products\Actions\Variants\CreateVariantGroupAction;
use App\Domain\Products\Actions\Variants\DeleteVariantGroupAction;
use App\Domain\Products\Actions\Variants\PatchVariantGroupAction;
use App\Domain\Products\Actions\Variants\ReplaceVariantGroupAction;
use App\Domain\Products\Models\Product;
use App\Domain\Products\Models\ProductPropertyValue;
use App\Domain\Products\Models\VariantGroup;
use Illuminate\Foundation\Testing\DatabaseTransactions;
use Illuminate\Validation\ValidationException;
use Tests\TestCase;

class VariantGroupActionsTest extends TestCase
{
    use DatabaseTransactions;

    const TEST_VALUE = 'clock';

    /** @var Product */
    private $product;

    /** @var Property */
    private $attribute;

    protected function setUp(): void
    {
        parent::setUp();

        $this->attribute = Property::factory()->createOne(['type' => PropertyType::STRING]);
        $this->product = Product::factory()->createOne();

        ProductPropertyValue::make($this->product, $this->attribute, self::TEST_VALUE)->save();
    }

    public function testCreateAction(): void
    {
        $fields = [
            'product_ids' => [$this->product->id],
            'attribute_ids' => [$this->attribute->id],
        ];

        $result = (new CreateVariantGroupAction())->execute($fields);

        $this->assertEquals($fields['product_ids'], $result->products->pluck('id')->toArray());
        $this->assertEquals($fields['attribute_ids'], $result->properties->pluck('id')->toArray());
    }

    public function testReplaceAction(): void
    {
        $group = $this->createVariantGroup();
        $attribute = Property::factory()->createOne();
        $product = Product::factory()->createOne();

        $fields = [
            'product_ids' => [$product->id],
            'attribute_ids' => [$attribute->id],
        ];

        $result = (new ReplaceVariantGroupAction())->execute($group->id, $fields);

        $this->assertEquals($fields['product_ids'], $result->products->pluck('id')->toArray());
        $this->assertEquals($fields['attribute_ids'], $result->properties->pluck('id')->toArray());
    }

    public function testPatchAction(): void
    {
        $group = $this->createVariantGroup();
        $product = Product::factory()->createOne();

        $fields = [
            'product_ids' => [$product->id],
        ];

        $result = (new PatchVariantGroupAction())->execute($group->id, $fields);

        $this->assertEquals($fields['product_ids'], $result->products->pluck('id')->toArray());
    }

    public function testDeleteAction(): void
    {
        $group = $this->createVariantGroup();

        (new DeleteVariantGroupAction())->execute($group->id);

        $this->assertEquals(0, Product::inVariantGroup($group->id)->count());
    }

    public function testValidation(): void
    {
        $group = $this->createVariantGroup();

        $product = Product::factory()->createOne();
        ProductPropertyValue::make($product, $this->attribute, self::TEST_VALUE)->save();

        $fields = [
            'product_ids' => array_merge($group->product_ids, [$product->id]),
            'attribute_ids' => $group->attribute_ids,
        ];

        $this->expectException(ValidationException::class);

        (new ReplaceVariantGroupAction())->execute($group->id, $fields);
    }

    private function createVariantGroup(): VariantGroup
    {
        $group = VariantGroup::create([
            'product_ids' => [$this->product->id],
            'attribute_ids' => [$this->attribute->id],
        ]);

        $this->product->update(['variant_group_id' => $group->id]);
        $group->properties()->attach($this->attribute->id);

        return $group;
    }
}
