<?php

namespace Tests\Integration\Domain\Products\Models;

use App\Domain\Categories\Models\Property;
use App\Domain\Categories\Models\PropertyDirectoryValue;
use App\Domain\Categories\Models\PropertyType;
use App\Domain\Products\Models\Product;
use App\Domain\Products\Models\ProductPropertyValue;
use Illuminate\Foundation\Testing\DatabaseTransactions;
use Illuminate\Support\Facades\Date;
use Tests\TestCase;

class ProductPropertyValueTest extends TestCase
{
    use DatabaseTransactions;

    /** @var Product */
    private $product;

    protected function setUp(): void
    {
        parent::setUp();
        $this->product = Product::factory()->createOne();
    }

    /**
     * @dataProvider provideValues
     */
    public function testSetValue(PropertyType $type, mixed $value): void
    {
        $property = Property::factory()->createOne(['type' => $type]);
        $propertyValue = new ProductPropertyValue(['product_id' => $this->product->id, 'property_id' => $property->id]);
        $fieldName = "value_{$type->value}";

        $propertyValue->setValue($type, $value);
        $propertyValue->save();

        $this->assertEquals($value, ProductPropertyValue::find($propertyValue->getKey())->{$fieldName});
    }

    /**
     * @dataProvider provideValues
     */
    public function testGetValue(PropertyType $type, mixed $expected): void
    {
        $property = Property::factory()->createOne(['type' => $type]);
        $fieldName = "value_{$type->value}";
        $propertyValue = ProductPropertyValue::create([
            'product_id' => $this->product->id,
            'property_id' => $property->id,
            'type' => $type,
            $fieldName => $expected,
        ]);

        $this->assertEquals($expected, ProductPropertyValue::find($propertyValue->getKey())->value($type));
    }

    public function provideValues(): array
    {
        return [
            'datetime' => [PropertyType::DATETIME(), Date::now()->milliseconds(0)],
            'null datetime' => [PropertyType::DATETIME(), null],
            'string' => [PropertyType::STRING(), 'Test string value'],
            'null string' => [PropertyType::STRING(), null],
            'integer' => [PropertyType::INTEGER(), 1256],
            'null integer' => [PropertyType::INTEGER(), null],
            'double' => [PropertyType::DOUBLE(), 58256.14],
            'null double' => [PropertyType::DOUBLE(), null],
            'link' => [PropertyType::LINK(), 'https://google.ru'],
            'null link' => [PropertyType::LINK(), null],
            'boolean' => [PropertyType::BOOLEAN(), true],
            'null boolean' => [PropertyType::BOOLEAN(), null],
        ];
    }

    public function testDirectoryValue(): void
    {
        $property = Property::factory()->createOne(['type' => PropertyType::DIRECTORY]);
        $directoryValue = PropertyDirectoryValue::factory()->forProperty($property)->createOne();

        $propertyValue = ProductPropertyValue::create([
            'product_id' => $this->product->id,
            'property_id' => $property->id,
            'type' => PropertyType::DIRECTORY,
            'value_directory' => $directoryValue->id,
        ]);

        $this->assertEquals(
            $directoryValue->id,
            ProductPropertyValue::find($propertyValue->getKey())->value(PropertyType::DIRECTORY())
        );
    }
}
