<?php

namespace Tests\Integration\Domain\Products\Models;

use App\Domain\Categories\Models\Category;
use App\Domain\Categories\Models\Property;
use App\Domain\Categories\Models\PropertyType;
use App\Domain\Products\Models\Product;
use App\Domain\Products\Models\ProductPropertyValue;
use Illuminate\Foundation\Testing\DatabaseTransactions;
use Tests\TestCase;

class ProductTest extends TestCase
{
    use DatabaseTransactions;

    public function testInCategoriesSelf(): void
    {
        $category = Category::factory()->createOne();
        $product = Product::factory()->inCategory($category)->createOne();

        $results = Product::inCategories($category->id)->get();

        $this->assertEquals([$product->id], $results->pluck('id')->toArray());
    }

    public function testInCategoriesDescendant(): void
    {
        $parent = Category::factory()->createOne();
        $child = Category::factory()->withParent($parent)->createOne();
        $product = Product::factory()->inCategory($child)->createOne();

        $results = Product::inCategories($parent->id)->get();

        $this->assertEquals([$product->id], $results->pluck('id')->toArray());
    }

    public function testInCategoriesMultiple(): void
    {
        $parent1 = Category::factory()->createOne();
        $child1 = Category::factory()->withParent($parent1)->createOne();
        $product1 = Product::factory()->inCategory($child1)->createOne();

        $parent2 = Category::factory()->createOne();
        $child2 = Category::factory()->withParent($parent2)->createOne();
        $product2 = Product::factory()->inCategory($child2)->createOne();

        $results = Product::inCategories([$parent1->id, $parent2->id])->get();

        $this->assertEqualsCanonicalizing(
            [$product1->id, $product2->id],
            $results->pluck('id')->toArray()
        );
    }

    public function testInCategoriesIgnoresSiblings(): void
    {
        $category = Category::factory()->createOne();
        $sibling = Category::factory()->createOne();
        Product::factory()->inCategory($sibling)->createOne();

        $results = Product::inCategories($category->id)->get();

        $this->assertEmpty($results);
    }

    /**
     * @dataProvider provideVariantValues
     */
    public function testVariantValues(array $attribute_values, string $expected): void
    {
        $product = Product::factory()->createOne();
        $attributes = collect();

        foreach ($attribute_values as $type => $values) {
            $attribute = Property::factory()->createOne(['type' => $type]);
            $attributes->push($attribute);

            shuffle($values);

            foreach ($values as $value) {
                ProductPropertyValue::make($product, $attribute, $value)->save();
            }
        }

        $this->assertEquals($expected, $product->variantValues($attributes));
    }

    public function provideVariantValues(): array
    {
        return [
            'no values' => [
                [PropertyType::INTEGER => []],
                '!@#$%'
            ],
            'exists one value' => [
                [PropertyType::INTEGER => [22]],
                '22'
            ],
            'exists multiple values' => [
                [PropertyType::INTEGER => [22, 54]],
                '22|54'
            ],
            'multiple attributes no values' => [
                [PropertyType::INTEGER => [], PropertyType::BOOLEAN => []],
                '!@#$%|!@#$%'
            ],
            'multiple attributes exists values' => [
                [PropertyType::INTEGER => [50, 34], PropertyType::STRING => ['value']],
                '34|50|value'
            ],
            'multiple attributes missing values' => [
                [PropertyType::INTEGER => [], PropertyType::STRING => ['value']],
                '!@#$%|value'
            ],
        ];
    }
}
