<?php

namespace Tests\Integration\Domain\Support\Actions;

use App\Domain\Support\Actions\BaseFileAction;
use App\Domain\Support\Models\Model;
use App\Support\Files\FileIntent;
use App\Support\Files\FileStorage;
use Closure;
use Illuminate\Database\Schema\Blueprint;
use Illuminate\Foundation\Testing\DatabaseTransactions;
use Illuminate\Support\Facades\Schema;
use Mockery;
use Tests\TestCase;

class BaseFileActionTest extends TestCase
{
    use DatabaseTransactions;

    /** @var BaseFileActionStub */
    private $testing;

    /** @var FileStorage|Mockery\LegacyMockInterface|Mockery\MockInterface */
    private $mockStorage;

    private string $fakePath;

    protected function setUp(): void
    {
        parent::setUp();

        $this->fakePath = "path/to/test1.png";
        Schema::create(BaseFileActionTestModel::TABLE_NAME, function (Blueprint $table) {
            $table->temporary();
            $table->id();
            $table->string('image')->nullable();
            $table->string('custom')->nullable();
        });

        $this->mockStorage = Mockery::mock(FileStorage::class);
        $this->testing = new BaseFileActionStub($this->mockStorage);
    }

    public function testAttach(): void
    {
        $intent = FileIntent::fake();
        $root = $this->createRoot();

        $this->mockStorage->shouldReceive('add')->andReturn($this->fakePath);

        $this->testing->executeAttach($root, $intent);

        $this->assertEquals($this->fakePath, $root->refresh()->image);
    }

    public function testAttachCustom(): void
    {
        $intent = FileIntent::fake();
        $root = $this->createRoot();

        $this->mockStorage->shouldReceive('add')->andReturn($this->fakePath);

        $callback = function (BaseFileActionTestModel $model, ?string $filePath) {
            $model->custom = $filePath;
            $model->save();
            return null;
        };

        $this->testing->executeAttach($root, $intent, $callback);

        $this->assertEquals($this->fakePath, $root->refresh()->custom);
    }

    public function testAttachExisting(): void
    {
        $intent = FileIntent::fake();
        $root = $this->createRoot("path/to/old.png");

        $this->mockStorage->shouldReceive('add')->andReturn($this->fakePath);
        $this->mockStorage->shouldReceive('remove')->once();

        $this->testing->executeAttach($root, $intent);

        $this->assertEquals($this->fakePath, $root->refresh()->image);
    }

    public function testDetach(): void
    {
        $root = $this->createRoot($this->fakePath);
        $this->mockStorage->shouldReceive('remove');

        $this->testing->executeDetach($root);

        $this->assertNull($root->refresh()->image);
    }

    public function testDetachCustom(): void
    {
        $root = $this->createRoot(null, $this->fakePath);

        $callback = function (BaseFileActionTestModel $model) {
            $filePath = $model->custom;
            $model->custom = null;
            $model->save();

            return $filePath;
        };

        $this->mockStorage->shouldReceive('remove')->once();

        $this->testing->executeDetach($root, $callback);

        $this->assertNull($root->refresh()->custom);
    }

    private function createRoot(?string $image = null, ?string $custom = null): BaseFileActionTestModel
    {
        $root = new BaseFileActionTestModel();
        $root->image = $image;
        $root->custom = $custom;
        $root->save();

        return $root;
    }
}

/**
 * @property string|null $image
 * @property string|null $custom
 */
class BaseFileActionTestModel extends Model
{
    const TABLE_NAME = 'test_base_file_action';

    protected $table = self::TABLE_NAME;
    protected static $unguarded = true;

    public $timestamps = false;
}

class BaseFileActionStub extends BaseFileAction
{
    protected function createModel(): Model
    {
        return new BaseFileActionTestModel();
    }

    /**
     * @param BaseFileActionTestModel $model
     * @param FileIntent $intent
     * @param string|Closure $target
     */
    public function executeAttach(BaseFileActionTestModel $model, FileIntent $intent, $target = 'image'): void
    {
        $this->attach($model->getKey(), $intent, $target);
    }

    /**
     * @param BaseFileActionTestModel $model
     * @param string|Closure $target
     */
    public function executeDetach(BaseFileActionTestModel $model, $target = 'image'): void
    {
        $this->detach($model->getKey(), $target);
    }
}
