<?php

namespace Tests\Integration\Domain\Support\Concerns;

use App\Domain\Support\Concerns\AppliesToAggregate;
use App\Domain\Support\Models\Model;
use Closure;
use Ensi\LaravelAuditing\Facades\Transaction;
use Illuminate\Database\Schema\Blueprint;
use Illuminate\Foundation\Testing\DatabaseTransactions;
use Illuminate\Support\Facades\Schema;
use RuntimeException;
use Tests\TestCase;

class AppliesToAggregateTest extends TestCase
{
    use DatabaseTransactions;

    const DEFAULT_ROOT_NAME = 'Root name';

    /** @var AppliesToAggregateTestAction */
    private $testing;

    protected function setUp(): void
    {
        parent::setUp();

        Schema::create(AppliesToAggregateRoot::TABLE_NAME, function (Blueprint $table) {
            $table->temporary();
            $table->id();
            $table->timestamps();
            $table->string('name')->nullable();
        });

        $this->testing = new AppliesToAggregateTestAction();
    }

    public function testApply(): void
    {
        $root = $this->createRoot();
        $action = function (AppliesToAggregateRoot $model) {
            return $model->name;
        };

        $result = $this->testing->executeApply($root, $action);

        $this->assertEquals(self::DEFAULT_ROOT_NAME, $result);
    }

    public function testApplyUsesCurrentTransaction(): void
    {
        $root = $this->createRoot();
        $action = function () {
            return Transaction::uid();
        };

        $result = $this->testing->executeApply($root, $action);

        $this->assertEquals(Transaction::uid(), $result);
    }

    public function testUpdateOrCreateNew(): void
    {
        $action = function (AppliesToAggregateRoot $model) {
            $model->name = 'testUpdateOrCreateNew';
        };

        $entity = $this->testing->executeUpdateOrCreate(null, $action);

        $this->assertDatabaseHas(
            AppliesToAggregateRoot::TABLE_NAME,
            ['id' => $entity->id, 'name' => 'testUpdateOrCreateNew']
        );
    }

    public function testUpdateOrCreateExisting(): void
    {
        $root = $this->createRoot();
        $action = function (AppliesToAggregateRoot $model) {
            $model->name = 'testUpdateOrCreateExisting';
        };

        $entity = $this->testing->executeUpdateOrCreate($root, $action);

        $this->assertEquals($root->getKey(), $entity->getKey());
        $this->assertDatabaseHas(
            AppliesToAggregateRoot::TABLE_NAME,
            ['id' => $entity->id, 'name' => 'testUpdateOrCreateExisting']
        );
    }

    public function testUpdateOrCreateNotSavesUnchanged(): void
    {
        $root = $this->createRoot();
        $this->travel(1)->minutes();

        $action = function (AppliesToAggregateRoot $model) {
            $model->name = self::DEFAULT_ROOT_NAME;
        };

        $entity = $this->testing->executeUpdateOrCreate($root, $action);

        $this->assertDatabaseHas(
            AppliesToAggregateRoot::TABLE_NAME,
            ['id' => $entity->id, 'name' => self::DEFAULT_ROOT_NAME, 'updated_at' => $root->updated_at]
        );
    }

    public function testDelete(): void
    {
        $root = $this->createRoot();

        $this->testing->executeDelete($root);

        $this->assertDatabaseMissing(AppliesToAggregateRoot::TABLE_NAME, ['id' => $root->getKey()]);
    }

    public function testDeleteCallsAction(): void
    {
        $root = $this->createRoot();
        $action = function () {
            throw new RuntimeException();
        };

        try {
            $this->testing->executeDelete($root, $action);
            $this->fail('Action was not called');
        } catch (RuntimeException $e) {
            //
        }

        $this->assertDatabaseHas(AppliesToAggregateRoot::TABLE_NAME, ['id' => $root->getKey()]);
    }

    private function createRoot(): AppliesToAggregateRoot
    {
        return AppliesToAggregateRoot::create(['name' => self::DEFAULT_ROOT_NAME]);
    }
}

/**
 * @property string $name
 */
class AppliesToAggregateRoot extends Model
{
    const TABLE_NAME = 'test_applies_to_aggregate_root';

    protected $table = self::TABLE_NAME;
    protected static $unguarded = true;
}


class AppliesToAggregateTestAction
{
    use AppliesToAggregate;

    protected function createModel(): Model
    {
        return new AppliesToAggregateRoot();
    }

    public function executeApply(AppliesToAggregateRoot $entity, Closure $action)
    {
        return $this->apply($entity->getKey(), $action);
    }

    public function executeUpdateOrCreate(?AppliesToAggregateRoot $entity, Closure $action): AppliesToAggregateRoot
    {
        $entityId = $entity !== null ? $entity->getKey() : null;

        return $this->updateOrCreate($entityId, $action);
    }

    public function executeDelete(AppliesToAggregateRoot $entity, ?Closure $action = null): void
    {
        $this->delete($entity->getKey(), $action);
    }
}
