<?php

namespace Tests\Integration\Domain\Support\Concerns;

use App\Domain\Support\Concerns\InteractsWithModels;
use App\Domain\Support\Models\Model;
use App\Exceptions\ConstraintViolationException;
use App\Exceptions\OperationRejectedException;
use Illuminate\Database\Eloquent\Relations\HasOne;
use Illuminate\Database\Eloquent\Relations\HasOneOrMany;
use Illuminate\Database\Schema\Blueprint;
use Illuminate\Foundation\Testing\DatabaseTransactions;
use Illuminate\Support\Facades\Schema;
use Tests\TestCase;

class InteractsWithModelsTest extends TestCase
{
    use DatabaseTransactions;

    /** @var InteractsWithModelsAction */
    private $testing;

    protected function setUp(): void
    {
        parent::setUp();

        Schema::create(InteractsWithModelsRoot::TABLE_NAME, function (Blueprint $table) {
            $table->temporary();
            $table->id();
            $table->string('name');

            $table->unique('name');
        });

        Schema::create(InteractsWithModelsRelated::TABLE_NAME, function (Blueprint $table) {
            $table->temporary();
            $table->id();
            $table->string('name');
            $table->foreignId('root_id')->nullable();
        });

        $this->testing = new InteractsWithModelsAction();
    }

    public function testSave(): void
    {
        $root = new InteractsWithModelsRoot(['name' => 'Test']);

        $this->testing->save($root);

        $this->assertGreaterThan(0, $root->getKey());
    }

    public function testSaveFailed(): void
    {
        $root = new InteractsWithModelsRoot(['name' => 'Test']);
        $root->eventResult = false;

        $this->expectException(OperationRejectedException::class);

        $this->testing->save($root);
    }

    public function testDelete(): void
    {
        $root = InteractsWithModelsRoot::create(['name' => 'Test']);

        $this->testing->delete($root);

        $this->assertFalse($root->exists);
    }

    public function testDeleteFailed(): void
    {
        $root = InteractsWithModelsRoot::create(['name' => 'Test']);
        $root->eventResult = false;

        $this->expectException(OperationRejectedException::class);

        $this->testing->delete($root);
    }

    public function testSaveRelated(): void
    {
        $root = InteractsWithModelsRoot::create(['name' => 'Test']);
        $related = new InteractsWithModelsRelated(['name' => 'One']);

        $this->testing->saveRelated($root->related(), $related);

        $this->assertGreaterThan(0, $related->getKey());
    }

    public function testSaveRelatedFailed(): void
    {
        $root = InteractsWithModelsRoot::create(['name' => 'Test']);
        $related = new InteractsWithModelsRelated(['name' => 'One']);
        $related->eventResult = false;

        $this->expectException(OperationRejectedException::class);

        $this->testing->saveRelated($root->related(), $related);
    }

    public function testConstraintUnique(): void
    {
        InteractsWithModelsRoot::create(['name' => 'Test']);
        $root = new InteractsWithModelsRoot(['name' => 'Test']);

        $this->testing->save($root);

        $this->assertNotNull($this->testing->constraintException);
        $this->assertTrue($this->testing->constraintException->isUniqueViolation());
    }

    public function testConstraintNotNull(): void
    {
        $root = new InteractsWithModelsRoot();

        $this->testing->save($root);

        $this->assertNotNull($this->testing->constraintException);
        $this->assertTrue($this->testing->constraintException->isNotNullViolation());
    }
}

class InteractsWithModelsRoot extends Model
{
    const TABLE_NAME = 'test_interacts_with_models_root';

    protected $table = self::TABLE_NAME;
    public $timestamps = false;
    protected static $unguarded = true;

    public $eventResult = true;

    public function related(): HasOne
    {
        return $this->hasOne(InteractsWithModelsRelated::class, 'root_id');
    }

    protected static function boot()
    {
        parent::boot();

        static::saving(function (InteractsWithModelsRoot $model) {
            return $model->eventResult;
        });

        static::deleting(function (InteractsWithModelsRoot $model) {
            return $model->eventResult;
        });
    }
}

class InteractsWithModelsRelated extends Model
{
    const TABLE_NAME = 'test_interacts_with_models_related';

    protected $table = self::TABLE_NAME;
    public $timestamps = false;
    protected static $unguarded = true;

    public $eventResult = true;

    protected static function boot()
    {
        parent::boot();

        static::saving(function (InteractsWithModelsRelated $model) {
            return $model->eventResult;
        });

        static::deleting(function (InteractsWithModelsRelated $model) {
            return $model->eventResult;
        });
    }
}

class InteractsWithModelsAction
{
    use InteractsWithModels;

    /** @var ConstraintViolationException|null */
    public $constraintException;

    public function save(Model $model): void
    {
        $this->saveOrThrow($model);
    }

    public function delete(Model $model): void
    {
        $this->deleteOrThrow($model);
    }

    public function saveRelated(HasOneOrMany $relation, Model $model): void
    {
        $this->saveRelatedOrThrow($relation, $model);
    }

    protected function handleConstraintViolation(ConstraintViolationException $exception): void
    {
        $this->constraintException = $exception;
    }
}
