<?php

namespace Tests\Integration\Elastic\Products\Actions;

use App\Core\Elastic\Product\Indexing\Actions\CreateOffersAction;
use Ensi\OffersClient\Api\OffersApi;
use Ensi\OffersClient\Dto\Offer;
use Ensi\OffersClient\Dto\OfferSaleStatusEnum;
use Ensi\OffersClient\Dto\SearchOffersResponse;
use Illuminate\Support\Collection;
use Mockery;
use Mockery\LegacyMockInterface;
use Mockery\MockInterface;

class CreateOffersActionTest extends OffersActionTestCase
{
    private OffersApi|LegacyMockInterface|MockInterface $mockApi;

    private CreateOffersAction $testing;

    protected function setUp(): void
    {
        parent::setUp();

        $this->mockApi = Mockery::mock(OffersApi::class);
        $this->testing = new CreateOffersAction($this->mockApi);
    }

    public function testNoOffers(): void
    {
        $this->mockSearchResult([]);
        $tickets = new Collection([$this->createProductTicket(1)]);

        $results = $this->testing->execute($tickets);

        $this->assertCount(0, $results);
    }

    public function testAddOfferInfo(): void
    {
        $productId = 10;
        $offer = new Offer([
            'id' => 10,
            'product_id' => $productId,
            'seller_id' => 15,
            'storage_address' => 'Верхняя полка',
            'sale_status' => OfferSaleStatusEnum::ON_SALE,
        ]);

        $this->mockSearchResult([$offer]);

        $result = $this->executeOne($productId);

        $result->assertActive();
        $result->assertSaleActive();

        $this->assertEquals(15, $result->sellerId);
        $this->assertEquals(10, $result->offerId);
        $this->assertEquals('Верхняя полка', $result->storageAddress);
    }

    public function testDeactivate(): void
    {
        $productId = 121;
        $offer = new Offer([
            'id' => 10,
            'product_id' => $productId,
            'seller_id' => 15,
            'storage_address' => 'Верхняя полка',
            'sale_status' => OfferSaleStatusEnum::OUT_SALE,
        ]);

        $this->mockSearchResult([$offer]);

        $result = $this->executeOne($productId);

        $result->assertActive();
        $result->assertNotSaleActive();
    }

    public function testAddBasePrice(): void
    {
        $productId = 100;
        $offer = new Offer([
            'id' => 10,
            'product_id' => $productId,
            'seller_id' => 15,
            'storage_address' => 'Верхняя полка',
            'sale_status' => OfferSaleStatusEnum::OUT_SALE,
            'base_price' => 150.0,
        ]);

        $this->mockSearchResult([$offer]);

        $result = $this->executeOne($productId);

        $result->assertPriceEquals(150);
    }

    private function mockSearchResult(array $offers): void
    {
        $this->mockApi->allows('searchOffers')
            ->andReturn(new SearchOffersResponse(['data' => $offers]));
    }

    private function executeOne(int $productId): TestOfferTicket
    {
        $tickets = new Collection([$this->createProductTicket($productId)]);

        return new TestOfferTicket($this->testing->execute($tickets)->first());
    }
}
