<?php

namespace Tests\Integration\Elastic\Products;

use App\Core\Elastic\Product\Indexing\PipelineIndexer;
use App\Core\Elastic\Product\Indexing\ProductsPipeline;
use App\Core\Elastic\Product\Indexing\ProductTicket;
use App\Core\Elastic\Product\ProductIndex;
use App\Domain\Products\Models\Product;
use Illuminate\Foundation\Testing\DatabaseTransactions;
use Illuminate\Support\Collection;
use LogicException;
use Mockery;
use Mockery\LegacyMockInterface;
use Mockery\MockInterface;
use Tests\TestCase;

class PipelineIndexerTest extends TestCase
{
    use DatabaseTransactions;

    const MISSING_ID = 1_000_000;

    private ProductsPipeline|LegacyMockInterface|MockInterface $mockPipeline;
    private ProductIndex|LegacyMockInterface|MockInterface $mockIndex;

    private PipelineIndexer $testing;

    protected function setUp(): void
    {
        parent::setUp();

        $this->mockPipeline = Mockery::mock(ProductsPipeline::class);
        $this->mockIndex = Mockery::mock(ProductIndex::class);

        $this->testing = new PipelineIndexer($this->mockPipeline, $this->mockIndex);
    }

    public function testPipeline(): void
    {
        $this->mockIndex->allows('bulk');

        $this->mockPipeline->shouldReceive('execute')->once();

        $this->testing->preload([Product::factory()->createOne()->id]);
    }

    public function testPipelineMissing(): void
    {
        $this->mockIndex->allows('bulk');

        $this->mockPipeline->shouldNotReceive('execute');

        $this->testing->preload([self::MISSING_ID]);
    }

    public function testSave(): void
    {
        $this->mockPipeline->allows('execute');
        $this->mockIndex->shouldReceive('bulk')->once();

        $this->testing->preload([Product::factory()->createOne()->id]);
    }

    public function testSaveSkipsFailed(): void
    {
        $this->mockPipeline->allows('execute')
            ->andReturnUsing(function (Collection $tickets) {
                $tickets->each(fn(ProductTicket $ticket) => $ticket->isFailed = true);
            });

        $this->mockIndex->shouldNotReceive('bulk');

        $this->testing->preload([Product::factory()->createOne()->id]);
    }

    public function testSaveDeleted(): void
    {
        $this->mockPipeline->allows('execute');

        $this->mockIndex->shouldReceive('bulk')->once();

        $this->testing->preload([self::MISSING_ID]);
    }

    public function testExecuteWithEmptyIds(): void
    {
        $this->mockIndex->shouldNotReceive('bulk');
        $this->mockPipeline->shouldNotReceive('execute')->never();

        $this->testing->preload([]);
    }

    public function testIndexOnMissingTicket(): void
    {
        $this->assertFalse($this->testing->index(self::MISSING_ID));
    }

    public function testIndexOnProcessedTicket(): void
    {
        $product = Product::factory()->createOne();

        $this->mockPipeline->allows('execute');
        $this->mockIndex->allows('bulk');

        $this->testing->preload([$product->id]);

        $this->assertTrue($this->testing->index($product->id));
    }

    public function testIndexOnFailedTicket(): void
    {
        $product = Product::factory()->createOne();
        $error = new LogicException('Error indexing');

        $this->mockPipeline->allows('execute')
            ->andReturnUsing(fn(Collection $tickets) => $tickets->each(fn(ProductTicket $ticket) => $ticket->fail($error)));

        $this->mockIndex->allows('bulk');

        $this->testing->preload([$product->id]);

        $this->expectException(LogicException::class);
        $this->expectExceptionMessage($error->getMessage());

        $this->testing->index($product->id);
    }
}
