<?php

namespace Tests\Integration\Elastic\Products\Stages;

use App\Core\Elastic\Product\Indexing\Stages\ProductInfoStage;
use App\Domain\Categories\Models\Category;
use App\Domain\Classifiers\Models\Brand;
use App\Domain\Products\Models\Product;
use App\Domain\Products\Models\ProductArchiveType;

class ProductInfoStageTest extends StageTestCase
{
    public function testExecute(): void
    {
        $brand = Brand::factory()->createOne();
        $category = Category::factory()->createOne();
        $product = Product::factory()->inCategory($category)->withRelations(brand: $brand)->createOne();

        $result = $this->executeOne(new ProductInfoStage(), $product);

        $result->assertNotFailed();
        $result->assertActive();
    }

    public function testBrand(): void
    {
        $brand = Brand::factory()->createOne();
        $product = $this->createProduct(brand: $brand);

        $result = $this->executeOne(new ProductInfoStage(), $product);

        $this->assertEquals($brand->name, $result->brandName);
        $this->assertEquals($brand->code, $result->brandCode);
    }

    public function testBrandMissing(): void
    {
        $product = $this->createProduct(brand: false);

        $result = $this->executeOne(new ProductInfoStage(), $product);

        $result->assertNotFailed();
        $this->assertNull($result->brandCode);
    }

    public function testInArchive(): void
    {
        $product = $this->createProduct(['archive' => ProductArchiveType::ARCHIVE_REJECT]);

        $result = $this->executeOne(new ProductInfoStage(), $product);

        $result->assertNotActive();
        $this->assertEquals(ProductArchiveType::ARCHIVE_REJECT, $result->archive);
    }

    public function testCategoryCodes(): void
    {
        $categories = Category::factory()->createTree(3);
        $product = $this->createProduct(category: $categories->last());

        $result = $this->executeOne(new ProductInfoStage(), $product);

        $this->assertEqualsCanonicalizing(
            $categories->pluck('code')->all(),
            $result->categoryCodes
        );
    }

    public function testCategoryName(): void
    {
        /** @var Category $category */
        $category = Category::factory()->createTree()->last();
        $product = $this->createProduct(category: $category);

        $result = $this->executeOne(new ProductInfoStage(), $product);

        $this->assertEquals($category->name, $result->categoryName);
    }

    public function testCategoryMissing(): void
    {
        $product = $this->createProduct(category: false);

        $result = $this->executeOne(new ProductInfoStage(), $product);

        $result->assertNotActive();
    }

    private function createProduct(
        array $fields = [],
        Category|bool $category = true,
        Brand|bool $brand = true
    ): Product {
        $factory = Product::factory();

        if ($category !== false) {
            $factory = $factory->inCategory(is_bool($category) ? Category::factory()->createOne() : $category);
        }

        if (is_bool($brand)) {
            $brand = $brand ? Brand::factory()->createOne() : null;
        }

        return $factory->withRelations(brand: $brand)->createOne($fields);
    }
}
