<?php

namespace Tests\Unit\Support\Action;

use App\Support\Actions\ActionDecorator;
use App\Support\Actions\DecoratesAction;
use Tests\UnitTestCase;

class DecoratesActionTest extends UnitTestCase
{
    /** @var DecoratedActionStub */
    private $testing;

    protected function setUp(): void
    {
        parent::setUp();
        $this->testing = new DecoratedActionStub();
    }

    public function testWithoutDecorators(): void
    {
        $this->assertEquals("A&B", $this->testing->execute('A', 'B'));
    }

    public function testDecorate(): void
    {
        $this->assertEquals(
            'A&C',
            $this->testing->replaceTwoOn('C')->execute('A', 'B')
        );
    }

    public function testDecorateResult(): void
    {
        $this->assertEquals(
            '(A&B)',
            $this->testing->wrapBrackets()->execute('A', 'B')
        );
    }

    public function testDecorateMulti(): void
    {
        $result = $this->testing
            ->replaceTwoOn('C')
            ->wrapBrackets()
            ->replaceOneOn('S')
            ->execute('A', 'B');

        $this->assertEquals('(S&C)', $result);
    }

    public function testDecorateField(): void
    {
        $this->assertEquals(
            'A&BS',
            $this->testing->withSuffix('S')->execute('A', 'B')
        );
        $this->assertEquals('', $this->testing->getSuffix());
    }

    public function testDecorateFieldFromParameters(): void
    {
        $this->assertEquals(
            '/1&B',
            $this->testing->withPrefix('P')->execute('1', 'B')
        );
    }

    public function testDecoratesFieldMulti(): void
    {
        $result = $this->testing->withPrefix('P')
            ->wrapBrackets()
            ->withSuffix('S')
            ->execute('A', 'B');

        $this->assertEquals('(PA&BS)', $result);
        $this->assertEquals('', $this->testing->getPrefix());
        $this->assertEquals('', $this->testing->getSuffix());
    }
}

class DecoratedActionStub
{
    use DecoratesAction;

    private string $suffix = '';
    private string $prefix = '';

    public function execute(string $one, string $two): string
    {
        return "$this->prefix$one&$two$this->suffix";
    }

    public function getSuffix(): string
    {
        return $this->suffix;
    }

    public function getPrefix(): string
    {
        return $this->prefix;
    }

    public function replaceOneOn(string $value): static|ActionDecorator
    {
        return $this->decorateCall(function (callable $chain, string $one, string $two) use ($value) {
            return $chain($value, $two);
        });
    }

    public function replaceTwoOn(string $value): static|ActionDecorator
    {
        return $this->decorateCall(fn(callable $chain, string $one, string $two) => $chain($one, $value));
    }

    public function wrapBrackets(): static|ActionDecorator
    {
        return $this->decorateResult(fn(string $source) => "($source)");
    }

    public function withSuffix(string $value): static|ActionDecorator
    {
        return $this->decorateState(fn() => ['suffix' => $value]);
    }

    public function withPrefix(string $value): static|ActionDecorator
    {
        return $this->decorateState(function (string $one) use ($value) {
            return [
                'prefix' => $one == 'A' ? $value : '/',
            ];
        });
    }
}
