<?php

namespace Tests\Unit\Support;

use App\Support\Optional;
use Exception;
use RuntimeException;
use Tests\UnitTestCase;

/**
 * Тесты Optional.
 */
class OptionalTest extends UnitTestCase
{
    /**
     * @dataProvider provideOptionalFactory
     * @param callable $factory
     */
    public function testCreate(callable $factory): void
    {
        $this->assertEquals('Test', $factory('Test')->value());
    }

    /**
     * @dataProvider provideOptionalFactory
     * @param callable $factory
     */
    public function testGetNotEmptyReturnsValue(callable $factory): void
    {
        $this->assertEquals('Test', $factory('Test')->get());
    }

    /**
     * @dataProvider provideOptionalFactory
     * @param callable $factory
     */
    public function testGetEmptyThrowsException(callable $factory): void
    {
        $this->expectException(RuntimeException::class);
        $this->expectExceptionMessage(Optional::EXCEPTION_MESSAGE);

        $factory()->get();
    }

    /**
     * @dataProvider provideOptionalFactory
     * @param callable $factory
     */
    public function testGetOrThrowNotEmptyReturnsValue(callable $factory): void
    {
        $this->assertEquals(500, $factory(500)->getOrThrow());
    }

    /**
     * @dataProvider provideOptionalFactory
     * @param callable $factory
     */
    public function testGetOrThrowEmptyThrowsException(callable $factory): void
    {
        $this->expectException(RuntimeException::class);
        $this->expectExceptionMessage(Optional::EXCEPTION_MESSAGE);

        $factory()->getOrThrow();
    }

    /**
     * @dataProvider provideOptionalFactory
     * @param callable $factory
     */
    public function testIsEmpty(callable $factory): void
    {
        $this->assertTrue($factory()->isEmpty());
        $this->assertFalse($factory('Test')->isEmpty());
    }

    /**
     * @dataProvider provideOptionalFactory
     * @param callable $factory
     */
    public function testOrElse(callable $factory): void
    {
        $this->assertEquals('Test', $factory()->orElse('Test')->value());
        $this->assertEquals('Test', $factory()->orElse(fn () => 'Test')->value());
    }

    /**
     * @dataProvider provideOptionalFactory
     * @param callable $factory
     */
    public function testOrElseThrow(callable $factory): void
    {
        $this->expectException(RuntimeException::class);
        $this->expectExceptionMessage('Error 404 "None"');

        $factory()->orElseThrow(null, 'Error %d "%s"', 404, 'None');
    }

    /**
     * @dataProvider provideOptionalFactory
     * @param callable $factory
     */
    public function testDefaultExceptionClass(callable $factory): void
    {
        $this->expectException(OptionalTestException::class);
        $this->expectExceptionMessage(Optional::EXCEPTION_MESSAGE);

        $factory()
            ->withDefaultException(OptionalTestException::class)
            ->getOrThrow();
    }

    /**
     * @dataProvider provideOptionalFactory
     * @param callable $factory
     */
    public function testDefaultExceptionClassWithArgs(callable $factory): void
    {
        $this->expectException(OptionalTestArgsException::class);
        $this->expectExceptionMessage('Exception code: 10, details: error');

        $factory()
            ->withDefaultException(OptionalTestArgsException::class, 10, 'error')
            ->getOrThrow();
    }

    /**
     * @dataProvider provideOptionalFactory
     * @param callable $factory
     */
    public function testDefaultExceptionInstance(callable $factory): void
    {
        $message = 'Error test exception instance';
        $this->expectException(OptionalTestException::class);
        $this->expectExceptionMessage($message);

        $factory()
            ->withDefaultException(new OptionalTestException($message))
            ->orElseThrow();
    }

    /**
     * @dataProvider provideOptionalFactory
     * @param callable $factory
     */
    public function testMapToClass(callable $factory): void
    {
        $result = $factory('Source string')
            ->map(OptionalTestMapStub::class)
            ->get();

        $this->assertInstanceOf(OptionalTestMapStub::class, $result);
        $this->assertEquals('Source string', $result->name);
        $this->assertEquals(30, $result->age);
    }

    /**
     * @dataProvider provideOptionalFactory
     * @param callable $factory
     */
    public function mapToClassWithAdditionalArgs(callable $factory): void
    {
        $result = $factory('Source string')
            ->map(OptionalTestMapStub::class, 45)
            ->get();

        $this->assertEquals(30, $result->age);
    }

    /**
     * @dataProvider provideOptionalFactory
     * @param callable $factory
     */
    public function testMapEmptyDoesNothing(callable $factory): void
    {
        $result = $factory()->map(OptionalTestMapStub::class);
        $this->assertTrue($result->isEmpty());
    }

    /**
     * @dataProvider provideOptionalFactory
     * @param callable $factory
     */
    public function testMapWithCallable(callable $factory): void
    {
        $callback = function (string $source) {
            return 'From ' . $source;
        };

        $result = $factory('Source string')
            ->map($callback)
            ->get();

        $this->assertEquals('From Source string', $result);
    }

    public function provideOptionalFactory(): array
    {
        $factoryBaseOptional = function ($value = null) {
            return optional($value);
        };

        $factoryExtendedOptional = function ($value = null) {
            return Optional::create($value);
        };

        return [
            [$factoryBaseOptional],
            [$factoryExtendedOptional],
        ];
    }
}


class OptionalTestException extends RuntimeException
{
}


class OptionalTestArgsException extends Exception
{
    /**
     * OptionalTestException constructor.
     * @param int $code
     * @param string $details
     */
    public function __construct(int $code, string $details = '')
    {
        parent::__construct(sprintf('Exception code: %d, details: %s', $code, $details));
    }
}


class OptionalTestMapStub
{
    public $name;

    public $age;

    /**
     * OptionalTestMapStub constructor.
     * @param $name
     * @param int $age
     */
    public function __construct($name, $age = 30)
    {
        $this->name = $name;
        $this->age = $age;
    }
}
